#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

from datetime import datetime
from enum import Enum
from typing import Tuple, List, Dict, Optional, Any
from kgforge.core.archetypes.resolver import Resolver

from kgforge.core.commons.context import Context
from kgforge.core.commons.files import is_valid_url
from kgforge.core.commons.parser import _parse_type
from kgforge.core.commons.query_builder import QueryBuilder

class CategoryDataType(Enum):
    DATETIME = "datetime"
    NUMBER = "number"
    BOOLEAN = "boolean"
    LITERAL = "literal"


type_map = {
    datetime: CategoryDataType.DATETIME,
    str: CategoryDataType.LITERAL,
    bool: CategoryDataType.BOOLEAN,
    int: CategoryDataType.NUMBER,
    float: CategoryDataType.NUMBER,
    complex: CategoryDataType.NUMBER,
}

format_type = {
    CategoryDataType.DATETIME: lambda x: f'"{x}"^^xsd:dateTime',
    CategoryDataType.NUMBER: lambda x: x,
    CategoryDataType.LITERAL: lambda x: f'"{x}"',
    CategoryDataType.BOOLEAN: lambda x: "'true'^^xsd:boolean" if x is True else "'false'^^xsd:boolean",
}

sparql_operator_map = {
    "__lt__": "<",
    "__le__": "<=",
    "__eq__": "=",
    "__ne__": "!=",
    "__gt__": ">",
    "__ge__": ">=",
}

class SPARQLQueryBuilder(QueryBuilder):

    @staticmethod
    def build(
        schema: Dict,
        resolvers: Optional[List[Resolver]],
        context: Context,
        *filters,
        **params,
    ) -> Tuple[List, List]:

        statements = list()
        sparql_filters = list()
        for index, f in enumerate(filters):
            last_path = f.path[-1]
            try:
                last_term = context.terms[last_path]
            except KeyError:
                last_term = None
            if last_path in ["id", "@id"]:
                property_path = "/".join(f.path[:-1])
            elif last_path == "@type":
                minus_last_path = f.path[:-1]
                minus_last_path.append("type")
                property_path = "/".join(minus_last_path)
            else:
                property_path = "/".join(f.path)
            try:
                if (
                    last_path in ["type", "@type"]
                    or last_path in ["id", "@id"]
                    or (last_term is not None and last_term.type == "@id")
                ):
                    if f.operator == "__eq__":
                        statements.append(f"{property_path} {_box_value_as_full_iri(f.value)}")
                    elif f.operator == "__ne__":
                        statements.append(f"{property_path} ?v{index}")
                        sparql_filters.append(f"FILTER(?v{index} != {f.value})")
                    else:
                        raise NotImplementedError(
                            f"supported operators are '==' and '!=' when filtering by type or id."
                        )
                else:
                    parsed_type, parsed_value = _parse_type(f.value, parse_str=False)
                    value_type = type_map[parsed_type]
                    value = format_type[value_type](parsed_value if parsed_value else f.value)
                    if value_type is CategoryDataType.LITERAL:
                        if f.operator not in ["__eq__", "__ne__"]:
                            raise NotImplementedError(f"supported operators are '==' and '!=' when filtering with a str.")
                        statements.append(f"{property_path} ?v{index}")
                        sparql_filters.append(f"FILTER(?v{index} = {_box_value_as_full_iri(value)})")
                    else:
                        statements.append(f"{property_path} ?v{index}")
                        sparql_filters.append(
                            f"FILTER(?v{index} {sparql_operator_map[f.operator]} {_box_value_as_full_iri(value)})"
                        )
            except NotImplementedError as nie:
                raise ValueError(f"Operator '{sparql_operator_map[f.operator]}' is not supported with the value '{f.value}': {str(nie)}")
        return statements, sparql_filters

def _box_value_as_full_iri(value):
    return f"<{value}>" if is_valid_url(value) else value


#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.
from rdflib import Literal, XSD
from typing import Any

import dateutil
import datetime
from dateutil.parser import ParserError


def _parse_type(value: Any, parse_str:bool = False):
    _type = type(value)
    try:
        if _type == str and parse_str:
            # integer
            if value.isnumeric():
                return int, value
            # double
            if float(value):
                return float, value

    except ValueError as ve:
        pass
    except TypeError as te:
        pass
    try:
        # always parse str for datetime. TODO: find a better way of parsing datetime literal
        if _type == str and ("^^<http://www.w3.org/2001/XMLSchema#dateTime>" in value or "^^xsd:dateTime" in value):
            # Datetime => example="2011-04-09T20:00:00Z"^^<http://www.w3.org/2001/XMLSchema#dateTime>
            value_parts = value.split("^^")
            parsed_value = value_parts[0]
            return datetime.datetime, parsed_value
        else:
            return _type,value
    except Exception as pe:
        return _type, value



#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

import pickle
from pathlib import Path
from typing import Optional, Union, List

import numpy as np

from kgforge.core.commons.exceptions import ConfigurationError
from kgforge.specializations.resolvers.entity_linking.service.entity_linking_service import EntityLinkerService
from kgforge.specializations.resources import EntityLinkingCandidate


class EntityLinkerServiceSkLearn(EntityLinkerService):

    def __init__(self, kb, aliases, model, index):
        super().__init__(is_distance=True)
        self.kb = kb
        self.aliases = aliases
        self.model = model
        self.index = index

    def generate_candidates(self, mentions, target, mention_context, limit, bulk) \
            -> Optional[Union[EntityLinkingCandidate, List[EntityLinkingCandidate]]]:

        def _(d, i):
            alias, uid = self.aliases[int(i)]
            label, definition = self.kb[uid]
            return EntityLinkingCandidate(d, label=label, altLabel=alias, id=uid, definition=definition)

        mentions_index = [(i, str(mention)) for i, mention in enumerate(mentions)]
        mentions_labels = {str(mention) for mention in mentions}
        embeddings = self.model.transform(mentions_labels)

        distances, indexes = self.index.kneighbors(embeddings, limit)
        results = np.stack((distances, indexes), axis=2)
        i_res = {m: [_(d, i) for d, i in rs] for m, rs in zip(mentions_labels, results)}
        return [(m, i_res[m]) for i, m in mentions_index]

    @staticmethod
    def from_pretrained(dirpath: Path, filename):
        filepath = dirpath / filename
        if filepath.is_file():
            with filepath.open(mode='rb') as f:
                kb = pickle.load(f)
                aliases = pickle.load(f)
                model = pickle.load(f)
                index = pickle.load(f)
                return EntityLinkerServiceSkLearn(kb, aliases, model, index)
        else:
            raise ConfigurationError(f"{dirpath}/{filename} is not a valid file path")



#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.
from pathlib import Path
from typing import List, Dict, Any, Optional, Callable, Union

from kgforge.core.archetypes import Resolver
from kgforge.core.archetypes.resolver import _build_resolving_query
from kgforge.core.commons.execution import not_supported
from kgforge.core.commons.sparql_query_builder import SPARQLQueryBuilder
from kgforge.core.commons.strategies import ResolvingStrategy
from kgforge.specializations.mappers import DictionaryMapper
from kgforge.specializations.mappings import DictionaryMapping
from kgforge.specializations.resolvers.store_service import StoreService


class OntologyResolver(Resolver):

    def __init__(self, source: str, targets: List[Dict[str, Any]], result_resource_mapping: str,
                 **source_config) -> None:
        super().__init__(source,  targets, result_resource_mapping, **source_config)

    @property
    def mapping(self) -> Callable:
        return DictionaryMapping

    @property
    def mapper(self) -> Callable:
        return DictionaryMapper

    def _resolve(self, text: Union[str, List[str]], target: Optional[str], type: Optional[str],
                 strategy: ResolvingStrategy, resolving_context: Any, limit: Optional[int], threshold: Optional[float]) \
            -> Optional[List[Dict]]:

        if isinstance(text, list):
            not_supported(("text", list))
        # Use as default type owl:Class
        if type is None:
            type = "Class"

        properties_to_filter_with = ['label', 'notation', 'prefLabel', 'altLabel']        
        query_template = """
        CONSTRUCT {{
            ?id a ?type ;
            label ?label ;
            prefLabel ?prefLabel ;
            altLabel ?altLabel ;
            definition ?definition;
            subClassOf ?subClassOf ;
            isDefinedBy ?isDefinedBy ;
            notation ?notation
        }} WHERE {{
            GRAPH ?g {{
                ?id a ?type ;
                    label ?label ; 
                OPTIONAL {{
                ?id subClassOf ?subClassOf ;
                }}
                OPTIONAL {{
                ?id definition ?definition ;
                }}
                OPTIONAL {{
                ?id prefLabel ?prefLabel .
                }}
                OPTIONAL {{
                ?id altLabel ?altLabel .
                }}
                OPTIONAL {{
                ?id isDefinedBy ?isDefinedBy .
                }}     
                OPTIONAL {{
                ?id notation ?notation .
                }}    
                {{
                SELECT * WHERE {{
                    {{ {0} ; label ?label {1} }} UNION
                    {{ {0} ; notation ?notation {2} }} UNION
                    {{ {0} ; prefLabel ?prefLabel {3} }} UNION
                    {{ {0} ; altLabel ?altLabel {4} }}
                }} LIMIT {5}
                }}
            }}
        }}
        """
        filters = self.service.filters[target] if target in self.service.filters else None
        context = self.service.get_context(resolving_context, target, filters)
        query, strategy_dependant_limit = _build_resolving_query(text, query_template, self.service.deprecated_property, filters, strategy, type, properties_to_filter_with, context, SPARQLQueryBuilder, limit)
        expected_fields = properties_to_filter_with+["type",  "definition", "subClassOf", "isDefinedBy"]
        return self.service.perform_query(query, target, expected_fields, strategy_dependant_limit)

    def _is_target_valid(self, target) -> Optional[bool]:
        return self.service.validate_target(target)

    @staticmethod
    def _service_from_directory(dirpath: Path, targets: Dict[str,  Dict[str, Dict[str, str]]], **source_config) -> Any:
        not_supported()

    @staticmethod
    def _service_from_store(store: Callable, targets: Dict[str,  Dict[str, Dict[str, str]]], **store_config) -> StoreService:
        return StoreService(store, targets, **store_config)



#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.
from abc import ABC, abstractmethod

from typing import List, Optional, Union, Any, Dict

from kgforge.core.commons.strategies import ResolvingStrategy
from kgforge.core.resource import encode
from kgforge.specializations.resources.entity_linking_candidate import EntityLinkingCandidate


class EntityLinkerService(ABC):

    def __init__(self, is_distance: bool = False):
        self.is_distance = is_distance

    @abstractmethod
    def generate_candidates(self, mentions: List[str], target:str, mention_context: Any, limit: int, bulk: bool)\
            -> Optional[List[EntityLinkingCandidate]]:
        pass

    def rank_candidates(self, candidates: List[EntityLinkingCandidate], strategy: ResolvingStrategy, threshold: float,
                        mention: str = None, mention_context: Any = None) -> Optional[List[Dict]]:

        exact_match_score = 0 if self.is_distance else 1
        threshold_operator = "<=" if self.is_distance else ">="
        is_sorted_reversed = True if not self.is_distance else False
        if strategy == ResolvingStrategy.EXACT_MATCH:
            zeros = [x for x in candidates if x.score == exact_match_score]
            if zeros and len(zeros) > 0:
                return [encode(zeros[0])]
            else:
                return None
        elif strategy == ResolvingStrategy.BEST_MATCH:
            chosen = sorted(candidates, key=lambda x: x.score, reverse=is_sorted_reversed)[0]
            return [encode(chosen)] if eval(f"{chosen.score} {threshold_operator} {threshold}") else None
        else:
            mentions = sorted(candidates, key=lambda x: x.score, reverse=is_sorted_reversed)
            return [encode(mention) for mention in mentions if eval(f"{mention.score} {threshold_operator} {threshold}")]



#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.
import itertools

import requests

from typing import Callable, Dict, List, Optional, Union, Any

from kgforge.core import Resource
from kgforge.core.archetypes import Store
from kgforge.core.conversions.json import as_json
from kgforge.core.resource import encode
from kgforge.core.wrappings import Filter, FilterOperator
from kgforge.specializations.mappers import DictionaryMapper
from kgforge.specializations.mappings import DictionaryMapping
from kgforge.specializations.resolvers.entity_linking.service.entity_linking_service import (
    EntityLinkerService,
)
from kgforge.specializations.resources.entity_linking_candidate import (
    EntityLinkingCandidate,
)


class EntityLinkerElasticService(EntityLinkerService):
    def __init__(
        self,
        store: Callable,
        targets: Dict[str, str],
        encoder,
        result_resource_mapping,
        **store_config
    ):
        super().__init__(is_distance=False)
        self.sources: Dict[str, Store] = dict()
        for identifier in targets:
            bucket = targets[identifier]['bucket']
            store_config.update(bucket=bucket)
            self.sources[identifier] = store(**store_config)
        self.encoder = encoder
        self.result_mapping: Any = self.mapping.load(result_resource_mapping)

    @property
    def mapping(self) -> Callable:
        return DictionaryMapping

    @property
    def mapper(self) -> Callable:
        return DictionaryMapper

    def generate_candidates(
        self, mentions, target, mention_context, limit, bulk
    ) -> Optional[Union[EntityLinkingCandidate, List[EntityLinkingCandidate]]]:
        def _(d, resource):
            return EntityLinkingCandidate(d, **resource)

        mentions_index = [(i, str(mention)) for i, mention in enumerate(mentions)]
        mentions_labels = {str(mention) for mention in mentions}

        resources, scores = [], []
        for mention in mentions_labels:
            call_url = self.encoder.format(x=mention)
            embedding_object = requests.get(url=call_url)
            embedding = self.mapper().map(
                embedding_object.json(), self.result_mapping, None
            )
            if embedding is not None:
                embedding_json = encode(embedding)
                vector_field = list(embedding_json.keys())[0]
                mention_resources, mention_resources_scores = self._similar(
                    vector_field,embedding_json[vector_field], target, limit
                )
                resources.append(mention_resources)
                scores.append(mention_resources_scores)
        i_res = {
            m: [_(scores[j][i], resource) for i, resource in enumerate(rs)]
            for j, (m, rs) in enumerate(
                zip(itertools.cycle(mentions_labels), resources)
            )
        }
        return [(m, i_res[m]) for i, m in mentions_index if m in i_res]

    def _similar(self, vector_field, item_embedding, target, limit, offset=0):
        """
        Given a vector, find similar top [limit] resources, ranked by cosine similarity
        """
        embedding_filter = Filter(
            operator=FilterOperator.EQUAL.value,
            path=[vector_field],
            value=item_embedding,
        )

        resources = self.sources[target].search(
            None,
            embedding_filter,
            limit=limit,
            offset=offset,
            excludes=[vector_field],
            search_endpoint="elastic",
        )

        if len(resources) > 0:
            scores = [
                r._store_metadata._score
                for r in resources
                if hasattr(r._store_metadata, "_score")
            ]

            return (
                as_json(
                    resources,
                    expanded=False,
                    store_metadata=True,
                    model_context=None,
                    metadata_context=None,
                    context_resolver=None,
                ),
                scores,
            )
        else:
            return None



#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

import os
from typing import Callable, List, Union, Dict, Optional
from uuid import uuid4

import pytest
from pytest_bdd import given, parsers, then, when

from kgforge.core import Resource, KnowledgeGraphForge
from kgforge.core.commons.actions import Action
from kgforge.core.commons.context import Context
from kgforge.core.conversions.rdf import _merge_jsonld, Form
from kgforge.core.wrappings.dict import wrap_dict


def do(fun: Callable, data: Union[Resource, List[Resource]], *args) -> None:
    if isinstance(data, List) and all(isinstance(x, Resource) for x in data):
        for x in data:
            fun(x, *args)
    elif isinstance(data, Resource):
        fun(data, *args)
    else:
        raise TypeError("not a Resource nor a list of Resource")


# It seems using 'given' with both 'fixture' and 'target_fixture' does not work.
# This implies that (in)valid resources creation steps need their dedicated function.


# Resource(s) creation.


@pytest.fixture
def json_one():
    return {"id": "123", "type": "Type", "p1": "v1a", "p2": "v2a"}


@pytest.fixture
def json_list_one():
    return {
        "id": "678",
        "type": "Other",
        "p3": "v3c",
        "p4": [
            {"id": "123", "type": "Type", "p1": "v1a", "p2": "v2a"},
            {"id": "345", "type": "Type", "p1": "v1b", "p2": "v2b"},
        ],
    }


def resource(valid: bool, index: int = 0) -> Resource:
    rid = str(uuid4())
    r = Resource(type="Person", id=rid)
    if valid:
        r.name = f"resource {index}"
    return r


def resources(valid: bool) -> List[Resource]:
    r1 = resource(valid, 0)
    r2 = resource(valid, 1)
    return [r1, r2]


@given("A valid resource.", target_fixture="data")
def valid_resource():
    return resource(True)


@given("Valid resources.", target_fixture="data")
def valid_resources():
    return resources(True)


@given("An invalid resource.", target_fixture="data")
def invalid_resource():
    return resource(False)


@given("Invalid resources.", target_fixture="data")
def invalid_resources():
    return resources(False)


# Resource(s) modifications.


@when("I modify the resource.")
def modify_resource(data):
    data.name = "other"


# Resource(s) verifications.


@then(parsers.parse("The '{attr}' status of {} resource should be '{value}'."))
def check_resource_status(data, attr, value):
    def fun(x):
        assert str(getattr(x, attr)) == value

    do(fun, data)


# Action(s) verifications.


def check_report(capsys, rc, err, msg, op):
    out = capsys.readouterr().out[:-1]
    heads = {
        "": "",
        "s": "<count> 2\n",
    }
    head = heads[rc]
    tails = {
        None: f"False\n<error> {msg}",
        " not": "True",
    }
    tail = tails[err]
    assert out == f"{head}<action> {op}\n<succeeded> {tail}"


@then("I should be able to access the report of the action on a resource.")
def check_action(data):
    def fun(x):
        assert isinstance(x._last_action, Action)

    do(fun, data)


@then(parsers.parse("The report should say that the operation was '{value}'."))
def check_action_operation(data, value):
    def fun(x):
        assert x._last_action.operation == value

    do(fun, data)


@then(parsers.parse("The report should say that the operation success is '{value}'."))
def check_action_success(data, value):
    def fun(x):
        assert str(x._last_action.succeeded) == value

    do(fun, data)


@then(parsers.parse("The report should say that the error was '{value}'."))
def check_action_error(data, value):
    def fun(x):
        assert str(x._last_action.error) == value

    do(fun, data)


# Fixtures for Resource to JSON-LD conversion and vice versa


@pytest.fixture
def forge():
    config = {
        "Model": {
            "name": "DemoModel",
            "origin": "directory",
            "source": "tests/data/demo-model/",
        },
        "Store": {
            "name": "DemoStore",
        },
    }
    return KnowledgeGraphForge(config)


@pytest.fixture
def custom_context():
    return {
        "@context": {
            "@base": "http://example.org/",
            "foaf": "http://xmlns.com/foaf/0.1/",
            "Person": "foaf:Person",
            "name": "foaf:name",
        }
    }


@pytest.fixture
def bbn_deprecated_property() -> str:
    return "https://bluebrain.github.io/nexus/vocabulary/deprecated"


@pytest.fixture
def metadata_context() -> Context:
    document = {
        "deprecated": "https://store.net/vocabulary/deprecated",
        "version": "https://store.net/vocabulary/version",
    }
    return Context(document, "http://store.org/metadata.json")


@pytest.fixture
def metadata_data_compacted():
    return {"deprecated": False, "version": 1}


@pytest.fixture
def metadata_data_expanded():
    return {
        "https://store.net/vocabulary/deprecated": [{"@value": False}],
        "https://store.net/vocabulary/version": [{"@value": 1}],
    }


@pytest.fixture
def store_metadata_context():
    return {
        "meta": "https://store.net/vocabulary/",
        "deprecated": "meta:deprecated",
        "version": "meta:version",
    }


@pytest.fixture
def store_metadata_value(store_metadata_context, metadata_data_compacted):
    data = {"context": store_metadata_context}
    data.update(metadata_data_compacted)
    return data


@pytest.fixture
def person(custom_context):
    return Resource(context=custom_context, type=["Person", "Agent"], name="Jami Booth")


@pytest.fixture
def registered_person_custom_context(person, custom_context, store_metadata_value):
    person.id = "c51f4e4e-2b30-41f4-8f7c-aced85632b03"
    store_metadata_value["id"] = person.id
    person._store_metadata = wrap_dict(store_metadata_value)
    return person


@pytest.fixture
def organization(registered_person_custom_context, store_metadata_value):
    contribution = Resource(
        type=["Organization", "Agent"],
        name="Reichel Inc",
        founder=registered_person_custom_context,
    )
    return contribution


@pytest.fixture
def building():
    type_ = "Building"
    name = "The Empire State Building"
    description = "The Empire State Building is a 102-story landmark in New York City."
    image = "http://www.civil.usherbrooke.ca/cours/gci215a/empire-state-building.jpg"
    geo = {"latitude": "40.75"}
    return Resource(
        type=type_, name=name, description=description, image=image, geo=geo
    )


@pytest.fixture(scope="session")
def context_file_path():
    return os.sep.join((os.path.abspath("."), "tests/data/shacl-model/context.json"))


@pytest.fixture(scope="session")
def context_iri_file(context_file_path):
    return f"file://{context_file_path}"


@pytest.fixture(scope="session")
def context_iri():
    return f"http://example.org/context"


@pytest.fixture(scope="session")
def model_context(context_iri_file) -> Context:
    return Context(context_iri_file, context_iri_file)


@pytest.fixture(scope="session")
def model_prefixes():
    return {
        "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
        "prov": "http://www.w3.org/ns/prov#",
        "schema": "http://schema.org/",
        "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
        "skos": "http://www.w3.org/2004/02/skos/core#",
        "nsg": "https://neuroshapes.org/",
        "owl": "http://www.w3.org/2002/07/owl#",
        "xsd": "http://www.w3.org/2001/XMLSchema#",
        "foaf": "http://xmlns.com/foaf/0.1/",
    }


@pytest.fixture
def building_jsonld(metadata_context, metadata_data_compacted, metadata_data_expanded):
    def _make_jsonld_expanded(resource, store_metadata, context):
        data = dict()
        if hasattr(resource, "id"):
            data["@id"] = resource.id
        ctx = (
            Context(resource.context)
            if hasattr(resource, "context")
            else Context(context)
        )
        latitude_term = ctx.terms.get("latitude")
        latitude_node = {
            "@value": resource.geo["latitude"],
        }
        if latitude_term.type:
            latitude_node = {**{"@type": latitude_term.type}, **latitude_node}

        geo_expanded = {latitude_term.id: [latitude_node]}
        data.update(
            {
                "@type": [ctx.expand(resource.type)],
                ctx.expand("description"): [{"@value":resource.description}],
                ctx.expand("geo"): [geo_expanded],
                ctx.expand("image"): [{"@id": resource.image}],
                ctx.expand("name"): [{"@value":resource.name}]
            }
        )
        if store_metadata and resource._store_metadata is not None:
            data.update(metadata_data_expanded)
        return data

    def _make_jsonld_compacted(resource, store_metadata, context):
        data = dict()
        data_context = resource.context if hasattr(resource, "context") else context
        if store_metadata and resource._store_metadata is not None:
            metadata_context_output = (
                metadata_context.iri
                if metadata_context.is_http_iri()
                else metadata_context.document["@context"]
            )
            data["@context"] = _merge_jsonld(data_context, metadata_context_output)
        else:
            data["@context"] = data_context
        if hasattr(resource, "id"):
            data["@id"] = resource.id
        data.update(
            {
                "@type": resource.type,
                "description": resource.description,
                "geo": resource.geo,
                "image": resource.image,
                "name": resource.name,
            }
        )
        if store_metadata and resource._store_metadata is not None:
            data.update(metadata_data_compacted)
        return data

    def _make_jsonld(
        rsrc: Resource,
        form: str,
        store_metadata: bool,
        context: Optional[Union[Dict, List, str]],
    ):
        if form is Form.COMPACTED.value:
            return _make_jsonld_compacted(rsrc, store_metadata, context)
        elif form is Form.EXPANDED.value:
            return _make_jsonld_expanded(rsrc, store_metadata, context)

    return _make_jsonld


SCOPE = "terms"
MODEL = "DemoModel"
STORE = "DemoStore"
RESOLVER = "DemoResolver"


@pytest.fixture(
    params=[
        MODEL,
        f"{MODEL} from kgforge.specializations.models",
        f"{MODEL} from kgforge.specializations.models.demo_model",
    ]
)
def model(request):
    return request.param


@pytest.fixture(
    params=[
        STORE,
        f"{STORE} from kgforge.specializations.stores",
        f"{STORE} from kgforge.specializations.stores.demo_store",
    ]
)
def store(request):
    return request.param


@pytest.fixture(
    params=[
        RESOLVER,
        f"{RESOLVER} from kgforge.specializations.resolvers",
        f"{RESOLVER} from kgforge.specializations.resolvers.demo_resolver",
    ]
)
def resolver(request):
    return request.param


@pytest.fixture
def config(model, store, resolver):
    return {
        "Model": {
            "name": model,
            "origin": "directory",
            "source": "tests/data/demo-model/",
        },
        "Store": {
            "name": store,
            "versioned_id_template": "{x.id}?_version={x._store_metadata.version}",
        },
        "Resolvers": {
            SCOPE: [
                {
                    "resolver": resolver,
                    "origin": "directory",
                    "source": "tests/data/demo-resolver/",
                    "targets": [
                        {
                            "identifier": "sex",
                            "bucket": "sex.json",
                            "filters":[
                                {
                                "path": "type",
                                "value": "class"
                                },
                                {
                                "path": "label",
                                "value": "female"
                                }
                            ]
                        },
                    ],
                    "result_resource_mapping": "../../configurations/demo-resolver/term-to-resource-mapping.hjson",
                },
            ],
        },
    }


@pytest.fixture
def es_mapping_dict():
    return {
        "dynamic": True,
        "properties": {
            "@id": {"type": "keyword"},
            "@type": {"type": "keyword"},
            "annotation": {
                "properties": {
                    "hasBody": {
                        "properties": {
                            "label": {
                                "fields": {"keyword": {"type": "keyword"}},
                                "type": "text",
                            },
                            "prefLabel": {
                                "fields": {"keyword": {"type": "keyword"}},
                                "type": "text",
                            },
                        },
                        "type": "nested",
                    }
                },
                "type": "object",
            },
            "atlasRelease": {
                "properties": {
                    "@id": {
                        "fields": {"keyword": {"type": "keyword"}},
                        "type": "keyword",
                    }
                },
                "type": "object",
            },
            "brainLocation": {
                "properties": {
                    "atlasSpatialReferenceSystem": {
                        "properties": {
                            "@id": {
                                "fields": {"keyword": {"type": "keyword"}},
                                "type": "keyword",
                            }
                        },
                        "type": "object",
                    },
                    "brainRegion": {
                        "properties": {
                            "@id": {
                                "fields": {"keyword": {"type": "keyword"}},
                                "type": "keyword",
                            },
                            "label": {
                                "fields": {"keyword": {"type": "keyword"}},
                                "type": "text",
                            },
                        },
                        "type": "object",
                    },
                    "coordinatesInBrainAtlas": {
                        "properties": {
                            "valueX": {
                                "properties": {
                                    "@type": {"type": "keyword"},
                                    "@value": {
                                        "fields": {"keyword": {"type": "keyword"}},
                                        "type": "float",
                                    },
                                },
                                "type": "object",
                            },
                            "valueY": {
                                "properties": {
                                    "@type": {"type": "keyword"},
                                    "@value": {
                                        "fields": {"keyword": {"type": "keyword"}},
                                        "type": "float",
                                    },
                                },
                                "type": "object",
                            },
                            "valueZ": {
                                "properties": {
                                    "@type": {"type": "keyword"},
                                    "@value": {
                                        "fields": {"keyword": {"type": "keyword"}},
                                        "type": "float",
                                    },
                                },
                                "type": "object",
                            },
                        },
                        "type": "object",
                    },
                    "layer": {
                        "properties": {
                            "label": {
                                "type": "text",
                            }
                        },
                        "type": "object",
                    },
                },
                "type": "object",
            },
            "contribution": {
                "properties": {
                    "agent": {
                        "properties": {
                            "@id": {
                                "fields": {"keyword": {"type": "keyword"}},
                                "type": "keyword",
                            },
                            "@type": {
                                "fields": {"keyword": {"type": "keyword"}},
                                "type": "text",
                            },
                        },
                        "type": "nested",
                    }
                },
                "type": "nested",
            },
            "derivation": {
                "properties": {
                    "entity": {
                        "properties": {
                            "@type": {
                                "fields": {"keyword": {"type": "keyword"}},
                                "type": "text",
                            },
                            "name": {
                                "fields": {"keyword": {"type": "keyword"}},
                                "type": "text",
                            },
                        },
                        "type": "nested",
                    },
                    "a_dense_vector": {
                        "dims": 3,
                        "type": "dense_vector",
                    },
                },
                "type": "nested",
            },
            "description": {"fields": {"keyword": {"type": "keyword"}}, "type": "text"},
            "dimension": {"type": "nested"},
            "distribution": {
                "properties": {
                    "contentSize": {"type": "nested"},
                    "contentUrl": {"type": "keyword"},
                    "digest": {
                        "properties": {"value": {"type": "keyword"}},
                        "type": "nested",
                    },
                    "encodingFormat": {"type": "keyword"},
                },
                "type": "nested",
            },
            "generation": {"type": "nested"},
            "isRegisteredIn": {
                "properties": {"@id": {"type": "keyword"}},
                "type": "object",
            },
            "license": {
                "properties": {
                    "label": {
                        "fields": {"keyword": {"type": "keyword"}},
                        "type": "text",
                    }
                },
                "type": "object",
            },
            "name": {"fields": {"keyword": {"type": "keyword"}}, "type": "text"},
            "an_integer": {
                "fields": {"keyword": {"type": "keyword"}},
                "type": "integer",
            },
            "a_dense_vector": {
                "dims": 3,
                "type": "dense_vector",
            },
            "a_boolean": {
                "type": "boolean",
            },
            "objectOfStudy": {
                "properties": {"label": {"type": "text"}},
                "type": "object",
            },
            "parcellationOntology": {
                "properties": {"@id": {"type": "keyword"}},
                "type": "object",
            },
            "parcellationVolume": {
                "properties": {"@id": {"type": "keyword"}},
                "type": "object",
            },
            "recordMeasure": {"type": "nested"},
            "series": {
                "properties": {
                    "statistic": {
                        "fields": {"keyword": {"type": "keyword"}},
                        "type": "text",
                    },
                    "unitCode": {
                        "fields": {"keyword": {"type": "keyword"}},
                        "type": "text",
                    },
                    "a_float": {
                        "type": "float",
                    },
                },
                "type": "nested",
            },
            "spatialReferenceSystem": {
                "properties": {
                    "@id": {
                        "fields": {"keyword": {"type": "keyword"}},
                        "type": "keyword",
                    }
                },
                "type": "object",
            },
            "subject": {"type": "object"},
            "_createdBy": {"type": "keyword"},
            "_updatedBy": {"type": "keyword"},
        },
    }



# 
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

from typing import List, Union
from warnings import warn

from kgforge.core import Resource
from kgforge.core.commons.actions import LazyAction
from kgforge.core.commons.execution import catch, not_supported
from kgforge.core.forge import KnowledgeGraphForge


# POLICY _set() should be only called as the last statement to ensure atomicity in case of errors.


class Dataset(Resource):
    """An opinionated high-level class based on Resource to handle datasets."""

    _RESERVED = {"_forge", "add_parts", "add_distribution", "add_contribution", "add_generation",
                 "add_derivation", "add_invalidation", "add_files",
                 "download"} | Resource._RESERVED

    # No catching of exceptions so that no incomplete instance is created if an error occurs.
    # This is a best practice in Python for __init__().
    def __init__(self, forge: KnowledgeGraphForge, type: str = "Dataset", **properties) -> None:
        super().__init__(**properties)
        self._forge: KnowledgeGraphForge = forge
        self.type: str = type

    @catch
    def add_parts(self, resources: List[Resource], versioned: bool = True) -> None:
        """Make resources part of the dataset."""
        keep = ["id", "type", "name", "distribution.contentUrl"]
        parts = self._forge.reshape(resources, keep, versioned)
        _set(self, "hasPart", parts)

    @catch
    def add_distribution(self, path: str, content_type: str = None) -> None:
        # path: FilePath.
        """Add a downloadable form of the dataset."""
        action = self._forge.attach(path, content_type)
        _set(self, "distribution", action)

    @catch
    def add_contribution(self, resource: Union[str, Resource], versioned: bool = True, **kwargs) -> None:
        # resource: Union[str, Resource].
        """Add information on the contribution of an agent during the generation of the dataset."""

        keep = ["id", "type", "_store_metadata"]
        contribution = self._add_prov_property(resource, "Contribution", "agent", "Agent", keep, versioned,
                                             **kwargs)
        _set(self, "contribution", contribution)

    @catch
    def add_generation(self, resource: Union[str, Resource], versioned: bool = True, **kwargs) -> None:
        """Add information about the activity that generated the dataset."""

        keep = ["id", "type", "_store_metadata"]
        generation = self._add_prov_property(resource, "Generation", "activity", "Activity", keep, versioned,
                                             **kwargs)
        _set(self, "generation", generation)

    @catch
    def add_derivation(self, resource: Union[str, Resource], versioned: bool = True, **kwargs) -> None:
        """Add information about the resources from which the dataset was derived."""

        keep = ["id", "type", "name", "_store_metadata"]
        derivation = self._add_prov_property(resource, "Derivation", "entity", "Entity", keep, versioned,
                                             **kwargs)
        _set(self, "derivation", derivation)

    @catch
    def add_invalidation(self, resource: Union[str, Resource], versioned: bool = True, **kwargs) -> None:
        """Add information about the invalidation of the dataset."""

        keep = ["id", "type", "_store_metadata"]
        invalidation = self._add_prov_property(resource,"Invalidation","activity","Activity",keep,versioned, **kwargs)
        _set(self, "invalidation", invalidation)

    def _add_prov_property(self,resource, prov_type, reference_property, reference_type, keep, versioned, **kwargs):

        if versioned and isinstance(resource, str):
            not_supported((f"resource:str when versioned is {versioned}. Set 'versioned' to False when referencing a str", True))
        if isinstance(resource, str):
            reference = Resource(type=reference_type, id=resource)
        elif isinstance(resource, Resource):
            try:
                reference = self._forge.reshape(resource, keep, versioned)
            except AttributeError as ae:
                if '_rev' in str(ae) and versioned:
                    raise ValueError(f"Missing resource revision value to build a versioned ({versioned}) reference. "
                                     f"Provide a revision number to the resource (by registering it for example) or set 'versioned' argument to False if no versioned reference is needed.")
                else:
                    raise ae

        result = Resource(type=prov_type, **kwargs)
        result.__setattr__(reference_property,reference)
        return result

    @catch
    def add_files(self, path: str, content_type: str = None) -> None:
        # path: DirPath.
        """Add (different) files as parts of the dataset."""
        action = self._forge.attach(path, content_type)
        distribution = Resource(distribution=action)
        _set(self, "hasPart", distribution)
 
    @catch
    def download(self, path: str = ".", source: str = "distributions", follow: str = None, overwrite: bool = False, cross_bucket: bool = False,
                 content_type: str = None) -> None:
        """
        Download the distributions of the dataset or the files part of the dataset.
        :param path: where to output the downloaded files
        :param source: [Deprecated] aliases to instruct forge the json path to use to collect downloadable file urls. Supported values are:
              [distributions] corresponding to the json path "distribution.contentUrl" (default)
              [parts] corresponding to the json path "hasPart.distribution.contentUrl"
        :param follow: the json property path holding a URL to download the files. This argument is exclusive to the 'source' argument
        :param overwrite: whether to replace (True) and existing file with the same name or not (False)
        :param cross_bucket: instructs the configured store to whether download files beyond the configured bucket (True) or not (False)
        :param content_type: the content_type of the files to download
        """
        if source:
            warn(
                DeprecationWarning(
                    "the source argument is deprecated and will be removed in nexusforge 1.0.0. Use the 'follow' argument instead"
                )
            )
        if source and follow:
            warn(
                Warning(
                    f"both 'source' {source} and 'follow' {follow} arguments are provided. 'follow' argument will be used. Consider providing one of the two arguments."
                )
            )
        if follow is None:
            if source == "distributions":
                follow = "distribution.contentUrl"
            elif source == "parts":
                follow = "hasPart.distribution.contentUrl"
            elif source:
                raise ValueError("unrecognized source. Use 'distributions' if the downloadable url is the value of 'distribution.contentUrl' or 'parts' if the downloadable url is the value of 'hasPart.distribution.contentUrl'")
            else:
                raise ValueError(
                    "'source' or 'follow' argument should be provided."
                )
        self._forge.download(self, follow, path, overwrite, cross_bucket, content_type)

    @classmethod
    def from_resource(cls, forge: KnowledgeGraphForge, data: Union[Resource, List[Resource]],
                      store_metadata: bool = False):
        def _(d):
            resource_json = forge.as_json(d)
            dataset = cls(forge, **resource_json)
            if store_metadata:
                dataset._store_metadata = d._store_metadata
            return dataset
        return [_(d) for d in data] if isinstance(data, List) else _(data)


def _set(dataset: Dataset, attr: str, data: Union[Resource, List[Resource], LazyAction]) -> None:
    if hasattr(dataset, attr):
        value = getattr(dataset, attr)
        if isinstance(value, List):
            if isinstance(data, List):
                value.extend(data)
            else:
                value.append(data)
        else:
            if isinstance(data, List):
                new = [value, *data]
            else:
                new = [value, data]
            setattr(dataset, attr, new)
    else:
        setattr(dataset, attr, data)



# 
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.
from pathlib import Path
from typing import Any, Dict, List, Callable, Optional, Tuple

from kgentitylinkingsklearn import EntityLinkerServiceSkLearn
from kgforge.core.commons.actions import LazyAction
from kgforge.specializations.mappers import DictionaryMapper
from kgforge.specializations.mappings import DictionaryMapping
from kgforge.specializations.resolvers import EntityLinker


class EntityLinkerSkLearn(EntityLinker):

    def __init__(self, source: str, targets: List[Dict[str, Any]], result_resource_mapping: str,
                 **source_config) -> None:
        super().__init__(source, targets, result_resource_mapping, **source_config)

    @property
    def mapping(self) -> Callable:
        return DictionaryMapping

    @property
    def mapper(self) -> Callable:
        return DictionaryMapper

    @staticmethod
    def _service_from_directory(dirpath: Path, targets: Dict[str, Tuple[str, Dict[str, str]]], **source_config) -> Dict[str, LazyAction]:
        # FIXME: the same model is loaded multiple times if provided for multiple targets
        return {target: LazyAction(EntityLinkerServiceSkLearn.from_pretrained, dirpath, filename_filter) for target, filename_filter in
                targets.items()}

    def _is_target_valid(self, target: str) -> Optional[bool]:
        if target and self.service and target not in self.service:
            raise ValueError(f"Unknown target value: {target}. Supported targets are: {self.service.keys()}")
        else:
            return True



# 
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.
from pathlib import Path
from typing import Callable, Dict, List, Any, Optional, Tuple

from kgforge.core.commons.execution import not_supported
from kgforge.specializations.mappers import DictionaryMapper
from kgforge.specializations.mappings import DictionaryMapping
from kgforge.specializations.resolvers.entity_linking import EntityLinker
from kgforge.specializations.resolvers.entity_linking.service.entity_linking_elastic_service import EntityLinkerElasticService


class EntityLinkerElastic(EntityLinker):

    def __init__(self, source: str, targets: List[Dict[str, Any]], result_resource_mapping: str,
                 **source_config) -> None:
        super().__init__(source, targets, result_resource_mapping, **source_config)

    @property
    def mapping(self) -> Callable:
        return DictionaryMapping

    @property
    def mapper(self) -> Callable:
        return DictionaryMapper

    @staticmethod
    def _service_from_directory(dirpath: Path, targets: Dict[str, Tuple[str, Dict[str, str]]]) -> Any:
        not_supported()

    @staticmethod
    def _service_from_store(store: Callable, targets: Dict[str, Tuple[str, Dict[str, str]]], **store_config) -> EntityLinkerElasticService:
        encoder = store_config.pop("encoder")
        encoder_url = encoder["source"]
        encoder_result_resource_mapping = encoder["result_resource_mapping"]
        return EntityLinkerElasticService(store, targets, encoder_url, encoder_result_resource_mapping, **store_config)

    def _is_target_valid(self, target: str) -> Optional[bool]:
        if target and target not in self.service.sources:
            raise ValueError(f"Unknown target value: {target}. Supported targets are: {self.service.sources.keys()}")
        else:
            return True



#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

# Placeholder for the test suite for actions.
import pytest

from kgforge.core import Resource, KnowledgeGraphForge
from kgforge.core.archetypes.store import rewrite_sparql, _replace_in_sparql
from kgforge.core.commons.context import Context
from kgforge.core.commons.exceptions import DownloadingError, FreezingError, QueryingError
from kgforge.specializations.resources import Dataset
from kgforge.core.wrappings.dict import DictWrapper, wrap_dict
import json

context = {
    "@context": {
        "@vocab":"http://example.org/vocab/",
        "type": {
            "@id": "rdf:type",
            "@type": "@id"
        },
        "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
        "prov": "http://www.w3.org/ns/prov#",
        "schema": "http://schema.org/",
        "Person": {
            "@id": "schema:Person",
            "@type": "@id"
        },
        "Association": "prov:Association",
        "name": "schema:name",
        "agent": "prov:agent",
        "description": "http://schema.org/description",
    }
}


prefixes = {
    "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
    "prov": "http://www.w3.org/ns/prov#",
    "schema": "http://schema.org/",
}


prefixes_string = "\n".join([f"PREFIX {k}: <{v}>" for k, v in prefixes.items()])

form_store_metadata_combinations = [
    ("SELECT ?agent WHERE { <http://exaplpe.org/1234> agent ?agent }",
     "\nSELECT ?agent WHERE { <http://exaplpe.org/1234> prov:agent ?agent }"),
    ("SELECT ?agent WHERE { ?agent type ?v0 FILTER(?v0 != Person) }",
     "\nSELECT ?agent WHERE { ?agent rdf:type ?v0 FILTER(?v0 != schema:Person) }"),
    ("SELECT ?agent WHERE { ?agent type ?v0 FILTER (?v0 in (Person, Agent)) }",
     "\nSELECT ?agent WHERE { ?agent rdf:type ?v0 FILTER (?v0 in (schema:Person, :Agent)) }"),
    ("SELECT ?x ?name WHERE { ?x type Association ; agent/name ?name }",
     "\nSELECT ?x ?name WHERE { ?x rdf:type prov:Association ; prov:agent/schema:name ?name }"),
    ("SELECT ?name WHERE { ?x agent/name ?name FILTER regex(?name, \"^j\", \"i\") }",
     "\nSELECT ?name WHERE { ?x prov:agent/schema:name ?name FILTER regex(?name, \"^j\", \"i\") }"),
    ("SELECT ?x WHERE { <http://exaplpe.org/1234> description ?x }",
     "\nSELECT ?x WHERE { <http://exaplpe.org/1234> <http://schema.org/description> ?x }"),
    ("SELECT ?x WHERE { <http://exaplpe.org/1234> a TypeNotInContext }",
     "\nSELECT ?x WHERE { <http://exaplpe.org/1234> a :TypeNotInContext }"),
    ("SELECT ?x WHERE { <http://exaplpe.org/1234> a TypeNotInContext, AnotherNotTypeInContext, Person }",
     "\nSELECT ?x WHERE { <http://exaplpe.org/1234> a :TypeNotInContext, :AnotherNotTypeInContext,"
                       " schema:Person }"),
    ("SELECT ?x WHERE { ?id propertyNotInContext ?x }",
     "\nSELECT ?x WHERE { ?id :propertyNotInContext ?x }"),
    ("SELECT ?x WHERE { ?id propertyNotInContext/name/anotherPropertyNotInContext ?x }",
     "\nSELECT ?x WHERE { ?id :propertyNotInContext/schema:name/:anotherPropertyNotInContext ?x }"),
    ("SELECT DISTINCT ?x WHERE { ?id propertyNotInContext/name/anotherPropertyNotInContext ?x }",
     "\nSELECT DISTINCT ?x WHERE { ?id :propertyNotInContext/schema:name/:anotherPropertyNotInContext ?x }"),
    ("SELECT ?x WHERE { Graph ?g { ?id propertyNotInContext/name/anotherPropertyNotInContext ?x }}",
     "\nSELECT ?x WHERE { Graph ?g { ?id :propertyNotInContext/schema:name/:anotherPropertyNotInContext ?x }}"),
    ("SELECT * WHERE { <http://exaplpe.org/1234> a TypeNotInContext, AnotherNotTypeInContext, Person; deprecated false.}",
     "\nSELECT * WHERE { <http://exaplpe.org/1234> a :TypeNotInContext, :AnotherNotTypeInContext, schema:Person; <https://store.net/vocabulary/deprecated> false.}")
]


@pytest.mark.parametrize("query, expected", form_store_metadata_combinations)
def test_rewrite_sparql(query, expected, metadata_context):
    prefixes_string_vocab = "\n".join([prefixes_string, f"PREFIX : <http://example.org/vocab/>"])
    context_object = Context(document=context)
    result = rewrite_sparql(query, context_object, metadata_context=metadata_context)
    assert result == prefixes_string_vocab + expected


def test_rewrite_sparql_unknownterm_missing_vocab(custom_context, metadata_context):
    context_object = Context(document=custom_context)
    assert not context_object.has_vocab()
    with pytest.raises(QueryingError):
        query = "SELECT ?x WHERE { Graph ?g { ?id propertyNotInContext/name/anotherPropertyNotInContext ?x }}"
        rewrite_sparql(query, context_object, metadata_context)


def test_rewrite_sparql_missingvocab(custom_context, metadata_context):
    query = "SELECT ?name WHERE { <http://exaplpe.org/1234> name ?name }"
    expected = "PREFIX foaf: <http://xmlns.com/foaf/0.1/>\nSELECT ?name WHERE { <http://exaplpe.org/1234> foaf:name ?name }"
    context_object = Context(document=custom_context)
    result = rewrite_sparql(query, context_object, metadata_context)
    assert result == expected

replace_in_sparql_combinations = [
    ("SELECT ?agent WHERE { <http://exaplpe.org/1234> prov:agent ?agent }",
     "LIMIT", 3, 100, r" LIMIT \d+", True,
     "SELECT ?agent WHERE { <http://exaplpe.org/1234> prov:agent ?agent }  LIMIT 3"),
    ("SELECT ?agent WHERE { <http://exaplpe.org/1234> prov:agent ?agent }",
     "LIMIT", None, 100, r" LIMIT \d+", True,
     "SELECT ?agent WHERE { <http://exaplpe.org/1234> prov:agent ?agent }  LIMIT 100"),
    ("SELECT ?agent WHERE { <http://exaplpe.org/1234> prov:agent ?agent }",
     "LIMIT", None, None, r" LIMIT \d+", True,
     "SELECT ?agent WHERE { <http://exaplpe.org/1234> prov:agent ?agent }"),
    ("SELECT ?agent WHERE { <http://exaplpe.org/1234> prov:agent ?agent } LIMIT 10",
     "LIMIT", None, None, r" LIMIT \d+", True,
     "SELECT ?agent WHERE { <http://exaplpe.org/1234> prov:agent ?agent } LIMIT 10"),
    ("SELECT ?agent WHERE { <http://exaplpe.org/1234> prov:agent ?agent } LIMIT 10",
     "LIMIT", None, 100, r" LIMIT \d+", True,
     "SELECT ?agent WHERE { <http://exaplpe.org/1234> prov:agent ?agent } LIMIT 100"),
    ("SELECT ?agent WHERE { <http://exaplpe.org/1234> prov:agent ?agent } LIMIT 10",
     "LIMIT", None, 100, r" LIMIT \d+", False,
     "SELECT ?agent WHERE { <http://exaplpe.org/1234> prov:agent ?agent } LIMIT 10"),
    ("SELECT ?agent WHERE { <http://exaplpe.org/1234> prov:agent ?agent }",
         "OFFSET", 1, 0, r" OFFSET \d+", True,
         "SELECT ?agent WHERE { <http://exaplpe.org/1234> prov:agent ?agent }  OFFSET 1"),
    ("SELECT ?agent WHERE { <http://exaplpe.org/1234> prov:agent ?agent }",
         "OFFSET", None, 0, r" OFFSET \d+", True,
         "SELECT ?agent WHERE { <http://exaplpe.org/1234> prov:agent ?agent }"),
    ("SELECT ?agent WHERE { <http://exaplpe.org/1234> prov:agent ?agent } OFFSET 3",
         "OFFSET", None, 20, r" OFFSET \d+", True,
         "SELECT ?agent WHERE { <http://exaplpe.org/1234> prov:agent ?agent } OFFSET 20"),
    ("SELECT ?agent WHERE { <http://exaplpe.org/1234> prov:agent ?agent } LIMIT 10 OFFSET 3",
         "OFFSET", 5, None, r" OFFSET \d+", True,
         "SELECT ?agent WHERE { <http://exaplpe.org/1234> prov:agent ?agent } LIMIT 10 OFFSET 5")
]

@pytest.mark.parametrize("query, what, value, default_value, search_regex, replace_if_in_query, expected",
                         replace_in_sparql_combinations)
def test__replace_in_sparql(query, what, value, default_value, search_regex, replace_if_in_query, expected):
    result = _replace_in_sparql(query, what, value, default_value, search_regex, replace_if_in_query)
    assert result == expected


def test__replace_in_sparql_exception():
    with pytest.raises(QueryingError):
        query = "SELECT ?agent WHERE { <http://exaplpe.org/1234> prov:agent ?agent } LIMIT 10"
        _replace_in_sparql(query, what="LIMIT", value=10, default_value=None, search_regex=r" LIMIT \d+",
                           replace_if_in_query=False)


def test_download(config):
    simple = Resource(type="Experiment", url="file.gz")
    with pytest.raises(DownloadingError):
        forge = KnowledgeGraphForge(config)
        forge._store.download(simple, "fake.path", "./", overwrite=False, cross_bucket=False)

def test_freeze(config, store_metadata_value):

    forge = KnowledgeGraphForge(config, debug=True)
    derivation1 = Dataset(forge, type="Dataset", name="A derivation dataset")
    derivation1.id = "http://derivation1"
    derivation1._store_metadata = wrap_dict(store_metadata_value)

    generation1 = Dataset(forge, type="Dataset", name="A generation dataset")
    generation1.id = "http://generation1"
    generation1._store_metadata = wrap_dict(store_metadata_value)

    invalidation1 = Dataset(forge, type="Activity", name="An invalidation activity")
    invalidation1.id = "http://invalidation1"
    invalidation1._store_metadata = wrap_dict(store_metadata_value)

    contribution1 = Resource(type="Person", name="A contributor")
    contribution1.id = "http://contribution1"
    contribution1._store_metadata = wrap_dict(store_metadata_value)

    dataset = Dataset(forge, type="Dataset", name="A dataset")
    dataset._store_metadata = wrap_dict(store_metadata_value)
    dataset.add_derivation(derivation1, versioned=False)
    dataset.add_generation(generation1, versioned=False)
    dataset.add_invalidation(invalidation1, versioned=False)
    dataset.add_contribution(contribution1, versioned=False)

    expected_derivation = json.loads(json.dumps({"type":"Derivation", "entity":{"id": "http://derivation1",
                                                                                "type":"Dataset", "name":"A derivation dataset"}}))
    assert forge.as_json(dataset.derivation) == expected_derivation

    expected_generation = json.loads(json.dumps({"type": "Generation",
                                                 "activity": {"id": "http://generation1", "type": "Dataset"}}))
    assert forge.as_json(dataset.generation) == expected_generation

    expected_contribution = json.loads(json.dumps({"type": "Contribution",
                                                 "agent": {"id": "http://contribution1", "type": "Person"}}))
    assert forge.as_json(dataset.contribution) == expected_contribution

    expected_invalidation = json.loads(json.dumps({"type": "Invalidation",
                                                   "activity": {"id": "http://invalidation1", "type": "Activity"}}))
    assert forge.as_json(dataset.invalidation) == expected_invalidation

    dataset.id = "http://dataset"
    dataset._synchronized = True
    forge._store.freeze(dataset)
    assert dataset.id == "http://dataset?_version=1"
    assert dataset.derivation.entity.id == "http://derivation1?_version=1"
    assert dataset.generation.activity.id == "http://generation1?_version=1"
    assert dataset.contribution.agent.id == "http://contribution1?_version=1"
    assert dataset.invalidation.activity.id == "http://invalidation1?_version=1"



#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

import asyncio
import copy
import json
import re
from asyncio import Task
from collections import namedtuple
from copy import deepcopy
from enum import Enum
from typing import Callable, Dict, List, Optional, Union, Tuple
from urllib.error import URLError
from urllib.parse import quote_plus, urlparse

import nest_asyncio
import nexussdk as nexus
import requests
from aiohttp import ClientSession, hdrs
from numpy import nan
from requests import HTTPError

from kgforge.core import Resource
from kgforge.core.commons.actions import (
    Action,
    collect_lazy_actions,
    execute_lazy_actions,
    LazyAction,
)
from kgforge.core.commons.exceptions import ConfigurationError
from kgforge.core.commons.context import Context
from kgforge.core.conversions.rdf import (
    _from_jsonld_one,
    _remove_ld_keys,
    as_jsonld,
    recursive_resolve,
)
from kgforge.core.wrappings.dict import wrap_dict


class BatchAction(Enum):
    CREATE = "create"
    FETCH = "fetch"
    DEPRECATE = "deprecate"
    UPDATE = "update"
    TAG = "tag"
    UPLOAD = "upload"
    DOWNLOAD = "download"


BatchResult = namedtuple("BatchResult", ["resource", "response"])
BatchResults = List[BatchResult]


class Service:

    NEXUS_NAMESPACE_FALLBACK = "https://bluebrain.github.io/nexus/vocabulary/"
    NEXUS_CONTEXT_FALLBACK = "https://bluebrain.github.io/nexus/contexts/resource.json"
    NEXUS_CONTEXT_SOURCE_FALLBACK = (
        "https%3A%2F%2Fbluebrain.github.io%2Fnexus%2Fcontexts%2Fresource.json/source"
    )
    DEFAULT_SPARQL_INDEX_FALLBACK = f"{NEXUS_NAMESPACE_FALLBACK}defaultSparqlIndex"
    DEFAULT_ES_INDEX_FALLBACK = f"{NEXUS_NAMESPACE_FALLBACK}defaultElasticSearchIndex"
    DEPRECATED_PROPERTY_FALLBACK = f"{NEXUS_NAMESPACE_FALLBACK}deprecated"
    PROJECT_PROPERTY_FALLBACK = f"{NEXUS_NAMESPACE_FALLBACK}project"

    def __init__(
        self,
        endpoint: str,
        org: str,
        prj: str,
        token: str,
        model_context: Context,
        max_connection: int,
        searchendpoints: Dict,
        store_context: str,
        store_local_context: str,
        namespace: str,
        project_property: str,
        deprecated_property: bool,
        content_type: str,
        accept: str,
        files_upload_config: Dict,
        files_download_config: Dict,
        **params,
    ):
        nexus.config.set_environment(endpoint)
        self.endpoint = endpoint
        self.organisation = org
        self.project = prj
        self.model_context = model_context
        self.context_cache: Dict = dict()
        self.max_connection = max_connection
        self.params = copy.deepcopy(params)
        self.store_context = store_context
        self.store_local_context = store_local_context
        self.namespace = namespace
        self.project_property = project_property
        self.store_metadata_keys = ["_constrainedBy","_createdAt","_createdBy","_deprecated","_incoming",
                                   "_outgoing","_project","_rev","_schemaProject", "_self","_updatedAt","_updatedBy"]

        self.deprecated_property = deprecated_property
        self.revision_property = f"{self.namespace}rev"
        self.default_sparql_index = f"{self.namespace}defaultSparqlIndex"
        self.default_es_index = f"{self.namespace}defaultElasticSearchIndex"

        self.headers = {"Content-Type": content_type, "Accept": accept}

        sparql_config = (
            searchendpoints["sparql"]
            if searchendpoints and "sparql" in searchendpoints
            else None
        )
        elastic_config = (
            searchendpoints["elastic"]
            if searchendpoints and "elastic" in searchendpoints
            else None
        )

        self.headers_sparql = {
            "Content-Type": sparql_config["Content-Type"]
            if sparql_config and "Content-Type" in sparql_config
            else "text/plain",
            "Accept": sparql_config["Accept"]
            if sparql_config and "Accept" in sparql_config
            else "application/sparql-results+json",
        }
        self.headers_elastic = {
            "Content-Type": elastic_config["Content-Type"]
            if elastic_config and "Content-Type" in elastic_config
            else "application/json",
            "Accept": elastic_config["Accept"]
            if elastic_config and "Accept" in elastic_config
            else "application/json",
        }
        self.headers_upload = {
            "Accept": files_upload_config.pop("Accept"),
        }
        self.headers_download = {"Accept": files_download_config.pop("Accept")}

        if token is not None:
            nexus.config.set_token(token)
            self.headers["Authorization"] = "Bearer " + token
            self.headers_sparql["Authorization"] = "Bearer " + token
            self.headers_elastic["Authorization"] = "Bearer " + token
            self.headers_upload["Authorization"] = "Bearer " + token
            self.headers_download["Authorization"] = "Bearer " + token
        self.context = Context(self.get_project_context())

        self.url_base_files = "/".join((self.endpoint, "files"))
        self.url_files = "/".join(
            (self.url_base_files, quote_plus(org), quote_plus(prj))
        )
        self.url_resources = "/".join(
            (self.endpoint, "resources", quote_plus(org), quote_plus(prj))
        )
        self.url_resolver = "/".join(
            (self.endpoint, "resolvers", quote_plus(org), quote_plus(prj))
        )
        self.metadata_context = Context(
            recursive_resolve(
                self.store_context, self.resolve_context, already_loaded=[]
            ),
            store_context,
        )
        sparql_view = (
            sparql_config["endpoint"]
            if sparql_config and "endpoint" in sparql_config
            else self.default_sparql_index
        )
        elastic_view = (
            elastic_config["endpoint"]
            if elastic_config and "endpoint" in elastic_config
            else self.default_es_index
        )
        default_str_keyword_field = (
            elastic_config["default_str_keyword_field"]
            if elastic_config and "default_str_keyword_field" in elastic_config
            else None
        )

        es_mapping = (
            elastic_config["mapping"]
            if elastic_config and "mapping" in elastic_config
            else None
        )

        self.sparql_endpoint = dict()
        self.elastic_endpoint = dict()
        self.sparql_endpoint["endpoint"] = "/".join(
            (
                self.endpoint,
                "views",
                quote_plus(org),
                quote_plus(prj),
                quote_plus(sparql_view),
                "sparql",
            )
        )
        self.sparql_endpoint["type"] = "sparql"

        self.elastic_endpoint["endpoint"] = "/".join(
            (
                self.endpoint,
                "views",
                quote_plus(org),
                quote_plus(prj),
                quote_plus(elastic_view),
                "_search",
            )
        )
        self.elastic_endpoint["type"] = "elastic"
        self.elastic_endpoint["view"] = LazyAction(
            nexus.views.fetch,
            quote_plus(org),
            quote_plus(prj),
            es_mapping
            if es_mapping
            else elastic_view,  # Todo consider using Dict for es_mapping
        )
        self.elastic_endpoint["default_str_keyword_field"] = default_str_keyword_field

        # The following code is for async to work on jupyter notebooks
        try:
            asyncio.get_event_loop()
            nest_asyncio.apply()
        except RuntimeError:
            pass

    def get_project_context(self) -> Dict:
        project_data = nexus.projects.fetch(self.organisation, self.project)
        context = {"@base": project_data["base"], "@vocab": project_data["vocab"]}
        for mapping in project_data['apiMappings']:
            context[mapping['prefix']] = mapping['namespace']
        return context

    def resolve_context(self, iri: str, local_only: Optional[bool] = False) -> Dict:
        if iri in self.context_cache:
            return self.context_cache[iri]
        try:
            context_to_resolve = (
                self.store_local_context if iri == self.store_context else iri
            )
            url = "/".join((self.url_resolver, "_", quote_plus(context_to_resolve)))
            response = requests.get(url, headers=self.headers)
            response.raise_for_status()
            resource = response.json()
        except Exception as e:
            if local_only is False:
                try:
                    context = Context(context_to_resolve)
                except URLError:
                    raise ValueError(f"{context_to_resolve} is not resolvable")
                else:
                    document = context.document["@context"]
            else:
                raise ValueError(f"{context_to_resolve} is not resolvable")
        else:
            # Make sure context is not deprecated
            if '_deprecated' in resource and resource['_deprecated']:
                raise ConfigurationError(f"Context {context_to_resolve} exists but was deprecated")
            document = json.loads(json.dumps(resource["@context"]))
        if isinstance(document, list):
            if self.store_context in document:
                document.remove(self.store_context)
            if self.store_local_context in document:
                document.remove(self.store_local_context)
        self.context_cache.update({context_to_resolve: document})
        return document

    def batch_request(
        self,
        resources: List[Resource],
        action: BatchAction,
        callback: Callable,
        error_type: Callable,
        **kwargs,
    ) -> Tuple[BatchResults, BatchResults]:
        def create_tasks(
            semaphore, session, loop, data, batch_action, f_callback, error
        ):
            futures = []
            schema_id = kwargs.get("schema_id")
            schema_id = "_" if schema_id is None else quote_plus(schema_id)
            for resource in data:
                params = deepcopy(kwargs.get("params", {}))
                if batch_action == batch_action.CREATE:
                    context = self.model_context or self.context
                    payload = as_jsonld(
                        resource,
                        "compacted",
                        False,
                        model_context=context,
                        metadata_context=None,
                        context_resolver=self.resolve_context
                    )
                    url = f"{self.url_resources}/{schema_id}"
                    prepared_request = loop.create_task(
                        queue(
                            hdrs.METH_POST,
                            semaphore,
                            session,
                            url,
                            resource,
                            error,
                            payload,
                            params,
                        )
                    )

                if batch_action == batch_action.UPDATE:
                    url = "/".join(
                        (self.url_resources, schema_id, quote_plus(resource.id))
                    )
                    params["rev"] = resource._store_metadata._rev

                    payload = as_jsonld(
                        resource,
                        "compacted",
                        False,
                        model_context=self.model_context,
                        metadata_context=None,
                        context_resolver=self.resolve_context
                    )
                    prepared_request = loop.create_task(
                        queue(
                            hdrs.METH_PUT,
                            semaphore,
                            session,
                            url,
                            resource,
                            error,
                            payload,
                            params,
                        )
                    )

                if batch_action == batch_action.TAG:
                    url = "/".join(
                        (self.url_resources, "_", quote_plus(resource.id), "tags")
                    )
                    rev = resource._store_metadata._rev
                    params["rev"] = rev
                    payload = {"tag": kwargs.get("tag"), "rev": rev}
                    prepared_request = loop.create_task(
                        queue(
                            hdrs.METH_POST,
                            semaphore,
                            session,
                            url,
                            resource,
                            error,
                            payload,
                            params,
                        )
                    )

                if batch_action == batch_action.DEPRECATE:
                    url = "/".join((self.url_resources, "_", quote_plus(resource.id)))
                    params["rev"] = resource._store_metadata._rev
                    prepared_request = loop.create_task(
                        queue(
                            hdrs.METH_DELETE,
                            semaphore,
                            session,
                            url,
                            resource,
                            error,
                            params=params,
                        )
                    )

                if batch_action == BatchAction.FETCH:
                    resource_org, resource_prj = resource._project.split("/")[-2:]
                    resource_url = "/".join(
                        (
                            self.endpoint,
                            "resources",
                            quote_plus(resource_org),
                            quote_plus(resource_prj),
                        )
                    )
                    url = "/".join((resource_url, "_", quote_plus(resource.id)))
                    if hasattr(resource, "_rev"):
                        params["rev"] = resource._rev
                    retrieve_source = params.get("retrieve_source", False)
                    if retrieve_source:
                        url = "/".join((url, "source"))
                    params.pop("retrieve_source")
                    prepared_request = loop.create_task(
                        queue(hdrs.METH_GET, semaphore, session, url, resource, error, params=params)
                    )

                if f_callback:
                    prepared_request.add_done_callback(f_callback)
                futures.append(prepared_request)
            return futures

        async def queue(
            method,
            semaphore,
            session,
            url,
            resource,
            exception,
            payload=None,
            params=None,
        ):
            async with semaphore:
                return await request(
                    method, session, url, resource, payload, params, exception
                )

        async def request(method, session, url, resource, payload, params, exception):
            async with session.request(
                method,
                url,
                headers=self.headers,
                data=json.dumps(payload),
                params=params,
            ) as response:
                content = await response.json()
                if response.status < 400:
                    return BatchResult(resource, content)
                else:
                    msg = " ".join(re.findall("[A-Z][^A-Z]*", content["@type"])).lower()
                    error = exception(msg)
                    return BatchResult(resource, error)

        async def dispatch_action():
            semaphore = asyncio.Semaphore(self.max_connection)
            loop = asyncio.get_event_loop()
            async with ClientSession() as session:
                tasks = create_tasks(
                    semaphore, session, loop, resources, action, callback, error_type
                )
                return await asyncio.gather(*tasks)

        return asyncio.run(dispatch_action())

    def sync_metadata(self, resource: Resource, result: Dict) -> None:
        metadata = (
            {"id": resource.id}
            if hasattr(resource, "id")
            else (
                {"id": resource.__getattribute__("@id")}
                if hasattr(resource, "@id")
                else dict()
            )
        )
        keys = sorted(self.metadata_context.terms.keys())
        keys.extend(["_index", "_score", "id", "@id"])
        only_meta = {k: v for k, v in result.items() if k in keys}
        metadata.update(_remove_ld_keys(only_meta, self.metadata_context, False))
        if not hasattr(resource, "id") and not hasattr(resource, "@id"):
            resource.id= result.get("id", result.get("@id",None))
        resource._store_metadata = wrap_dict(metadata)

    def synchronize_resource(
        self,
        resource: Resource,
        response: Union[Exception, Dict],
        action_name: str,
        succeeded: bool,
        synchronized: bool,
    ) -> None:
        if succeeded:
            action = Action(action_name, succeeded, None)
            self.sync_metadata(resource, response)
        else:
            action = Action(action_name, succeeded, response)
        resource._last_action = action
        resource._synchronized = synchronized

    def default_callback(self, fun_name: str) -> Callable:
        def callback(task: Task):
            result = task.result()
            if isinstance(result.response, Exception):
                self.synchronize_resource(
                    result.resource, result.response, fun_name, False, False
                )
            else:
                self.synchronize_resource(
                    result.resource, result.response, fun_name, True, True
                )

        return callback

    def verify(
        self,
        resources: List[Resource],
        function_name,
        exception: Callable,
        id_required: bool,
        required_synchronized: bool,
        execute_actions: bool,
    ) -> List[Resource]:
        valid = list()
        for resource in resources:
            if id_required and not hasattr(resource, "id"):
                error = exception("resource should have an id")
                self.synchronize_resource(resource, error, function_name, False, False)
                continue
            if required_synchronized is not None:
                synchronized = resource._synchronized
                if synchronized is not required_synchronized:
                    be_or_not_be = "be" if required_synchronized is True else "not be"
                    error = exception(f"resource should {be_or_not_be} synchronized")
                    self.synchronize_resource(
                        resource, error, function_name, False, False
                    )
                    continue
            if execute_actions:
                lazy_actions = collect_lazy_actions(resource)
                if lazy_actions is not None:
                    try:
                        execute_lazy_actions(resource, lazy_actions)
                    except Exception as e:
                        self.synchronize_resource(
                            resource, exception(e), function_name, False, False
                        )
                        continue
            valid.append(resource)
        return valid

    def to_resource(
        self, payload: Dict, sync_metadata: bool = True, **kwargs
    ) -> Resource:
        # Use JSONLD context defined in Model if no context is retrieved from payload
        # Todo: BlueBrainNexus store is not indexing in ES the JSONLD context, user provided context can be changed to Model defined one
        data_context = deepcopy(payload.get("@context", self.model_context.iri if self.model_context else None))
        if not isinstance(data_context, list):
            data_context = [data_context]
        if self.store_context in data_context:
            data_context.remove(self.store_context)
        if self.store_local_context in data_context:
            data_context.remove(self.store_local_context)
        data_context = data_context[0] if len(data_context) == 1 else data_context
        metadata = dict()
        data = dict()
        for k, v in payload.items():
            if k in self.metadata_context.terms.keys():
                metadata[k] = v
            else:
                data[k] = v

        if (
            self.model_context
            and data_context is not None
            and data_context == self.model_context.iri
        ):
            resolved_ctx = self.model_context.document["@context"]
        elif data_context is not None:
            resolved_ctx = recursive_resolve(
                data_context,
                self.resolve_context,
                already_loaded=[self.store_local_context, self.store_context],
            )
        else:
            resolved_ctx = None
        if resolved_ctx:
            data["@context"] = resolved_ctx
            resource = _from_jsonld_one(data)
            resource.context = data_context
        else:
            resource = Resource.from_json(data)

        if len(metadata) > 0 and sync_metadata:
            metadata.update(kwargs)
            self.sync_metadata(resource, metadata)
        if not hasattr(resource, "id") and kwargs and 'id' in kwargs.keys():
            resource.id = kwargs.get("id")
        return resource



#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

import pytest

from kgforge.core.archetypes.resolver import escape_punctuation, write_sparql_filters


@pytest.mark.parametrize(
    "text,expected",
    [
    pytest.param(("My label"), ("My label"),
        id="nothing"),
    pytest.param(("My label-"), ("My label\\\\-"),
        id="dash"),
    pytest.param(("My/label"), ("My\\\\/label"),
        id="slash"),
    pytest.param(("My +label"), ("My \\\\+label"),
        id="plus"),
    pytest.param(("My {}label"), ("My \\\\{\\\\}label"),
        id="curly_brackets"),
    pytest.param(("My <label>"), ("My \\\\<label\\\>"),
        id="brackets"),
    pytest.param(("My@label yay"), ("My\\\\@label yay"),
        id="atsymbol"),
    pytest.param(("My label: something."), ("My label\\\\: something\\\\."),
        id="dots"),
    pytest.param(("My label?!"), ("My label\\\\?\\\\!"),
        id="marks"),
    pytest.param(("The rest, ~ | are here # ;"),
        ("The rest\\\\, \\\\~ \\\\| are here \\\\# \\\\;"),
        id="rest")
    ]
)
def test_escape_punctuation(text, expected):
    new_text = escape_punctuation(text)
    assert new_text == expected

@pytest.fixture
def text():
    return "My nice text"

@pytest.fixture
def check_text():
    return "Label/with_Different+ Symbols."

agent_properties = ["name", "givenName", "familyName"]
ontology_properties = ['label', 'notation', 'prefLabel', 'altLabel']

def test_write_filters_params(text):
    with pytest.raises(ValueError):
        write_sparql_filters(text, ['name'], case_insensitive=True)

@pytest.mark.parametrize(
    "properties,params,expected",
    [
    pytest.param((agent_properties), ({}),
        ([' FILTER (?name = "My nice text")',
          ' FILTER (?givenName = "My nice text")',
          ' FILTER (?familyName = "My nice text")'],
         ),
        id="agent filters",
        ),
    pytest.param((agent_properties), ({'regex': True}),
        ([' FILTER regex(?name, "My nice text")',
          ' FILTER regex(?givenName, "My nice text")',
          ' FILTER regex(?familyName, "My nice text")'],
         ),
        id="agent filters regex",
        ),
    pytest.param((agent_properties), ({'regex': True, 'case_insensitive': True}),
        ([' FILTER regex(?name, "My nice text", \"i\")',
          ' FILTER regex(?givenName, "My nice text", \"i\")',
          ' FILTER regex(?familyName, "My nice text", \"i\")'],
         ),
        id="agent filters all",
        ),
    pytest.param((ontology_properties), ({}),
        ([' FILTER (?label = "My nice text")',
          ' FILTER (?notation = "My nice text")',
          ' FILTER (?prefLabel = "My nice text")',
          ' FILTER (?altLabel = "My nice text")'],
         ),
        id="ontology filters",
        ),
    pytest.param((ontology_properties), ({'regex': True, 'case_insensitive': True}),
        ([' FILTER regex(?label, "My nice text", \"i\")',
          ' FILTER regex(?notation, "My nice text", \"i\")',
          ' FILTER regex(?prefLabel, "My nice text", \"i\")',
          ' FILTER regex(?altLabel, "My nice text", \"i\")'],
         ),
        id="ontology filters",
        ),
    ]
)
def test_write_sparql_filters(text, properties, params, expected):
    filters = write_sparql_filters(text, properties, **params)
    assert filters == expected[0]


#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.
from typing import Any

from kgforge.core import Resource


class EntityLinkingCandidate(Resource):

    def __init__(self, score, candidate_context:Any = None, **properties):
        self.score = score
        self.candidate_context = candidate_context
        super().__init__(**properties)

    def __repr__(self):
        attrs = (f"{k}={v!r}" for k, v in self.__dict__.items())
        return f"Candidate({', '.join(attrs)})"



# 
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.
from abc import ABCMeta
from typing import Dict, List, Optional, Tuple, Union, Any

from kgforge.core import Resource
from kgforge.core.archetypes import Resolver
from kgforge.core.commons.actions import LazyAction
from kgforge.core.commons.exceptions import ResolvingError
from kgforge.core.commons.execution import not_supported
from kgforge.core.commons.strategies import ResolvingStrategy


class EntityLinker(Resolver, metaclass=ABCMeta):

    def __init__(self, source: str, targets: List[Dict[str, Any]], result_resource_mapping: str,
                 **source_config) -> None:
        super().__init__(source, targets, result_resource_mapping, **source_config)

    def _resolve(self, text: Union[str, List[str]], target: str, type: str,
                 strategy: ResolvingStrategy, resolving_context: Any, limit: Optional[int],
                 threshold: Optional[float]) -> Optional[List[Tuple[str, List[Dict]]]]:
        if target is not None:
            if isinstance(self.service, dict):
                entity_linker_service = self.service[target]
            else:
                entity_linker_service = self.service
        else:
            raise ResolvingError("A target is required for Entity Linker.")
        if isinstance(entity_linker_service, LazyAction):
            entity_linker_service = entity_linker_service.execute()
            self.service[target] = entity_linker_service
        mentions = [text] if isinstance(text, str) else text
        candidates = entity_linker_service.generate_candidates(mentions=mentions, target=target, mention_context=resolving_context,
                                                               limit=limit, bulk=False)
        return [(m, entity_linker_service.rank_candidates(candidates=cl, strategy=strategy, threshold=threshold,
                                                          mention=m, mention_context=resolving_context)) for m, cl in candidates]



# 
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

import pytest
from pytest_bdd import given, scenarios, then, when

from kgforge.core import Resource

# TODO To be port to the generic parameterizable test suite for resources in test_resources.py.
#  DKE-135.
from kgforge.core.resource import encode

scenarios("resource.feature")


@given("I create a resource with a property.")
def resource():
    return Resource(type="Entity")


@given("I create a resource from a JSON dict.")
def resource_from_json(json_one):
    return Resource.from_json(json_one)


@given("I create a resource from a JSON dict and exclude fields.")
def resource_from_json_na_one(json_one):
    return Resource.from_json(json_one, na="v1a")


@given("I create a resource from a JSON dict and exclude fields.")
def resource_from_json_na_many(json_one):
    return Resource.from_json(json_one, na=["v1a","v2a"])


@given("I create a resource with an other resource as property.")
def nresource():
    return Resource(type="Entity", contribution=Resource(type="Contribution"),
                    used=[Resource(type="Entity"), Resource(type="Dataset")])

@given("I create a resource with an id and type.")
def resource_id_type():
    return Resource(id="https://id", type="Person", name="name")


@when("I create a resource with a reserved attribute. Creation should fail.")
def reserved_attribute_error():
    with pytest.raises(NotImplementedError):
        Resource(_validated=True)


@then("I should be able to access properties as object attribute.")
def access_property(resource, resource_from_json, resource_from_json_na_one, resource_from_json_na_many):
    assert resource.type == "Entity"
    assert list(encode(resource_from_json).keys()) == ["id","type","p1","p2"]
    assert list(encode(resource_from_json).values()) == ["123","Type","v1a","v2a"]
    assert list(encode(resource_from_json_na_one).keys()) == ["id","type","p2"]
    assert list(encode(resource_from_json_na_one).values()) == ["123","Type","v2a"]
    assert list(encode(resource_from_json_na_many).keys()) == ["id","type"]
    assert list(encode(resource_from_json_na_many).values()) == ["123","Type"]


@then("I should be able to access the nested resource properties as JSONPath.")
def access_nested_property(nresource):
    assert nresource.type == "Entity"
    assert nresource.contribution.type == "Contribution"
    assert nresource.used[0].type == "Entity"
    assert nresource.used[1].type == "Dataset"


@then("I assigned _synchronized to True so the resource should give _inner_sync equals True.")
def change_property(resource):
    resource._synchronized = True
    assert resource._inner_sync == True
    resource.type = "test"
    assert resource._synchronized == False


@then("I changed a nested property so the resource should give _synchronized equals False.")
def change_nested_property(nresource):
    nresource._synchronized = True 
    assert nresource._inner_sync == True
    # Change some property
    nresource.contribution.type = "test"
    assert nresource._synchronized == False
    assert nresource._inner_sync == False
    # Synchronize again
    nresource._synchronized = True 
    assert nresource._inner_sync == True
    # Change a list
    for u in nresource.used:
        u.value = 10
    assert nresource._synchronized == False
    assert nresource._inner_sync == False

@then("I should be able to get its type and identifier.")
def change_nested_property(resource_id_type):
    assert resource_id_type.has_identifier(return_attribute=True) == (True, "id")
    assert resource_id_type.has_identifier() is True
    assert resource_id_type.has_type(return_attribute=True) == (True, "type")
    assert resource_id_type.has_type() is True
    
    assert resource_id_type.get_identifier() == "https://id"
    assert resource_id_type.get_type() == "Person"

    resource_no_id_no_type = Resource(name="name")
    assert resource_no_id_no_type.has_identifier(return_attribute=True) == (False, None)
    assert resource_no_id_no_type.has_identifier() is False
    assert resource_no_id_no_type.has_type(return_attribute=True) == (False, None)
    assert resource_no_id_no_type.has_type() is False
    
    assert resource_no_id_no_type.get_identifier() is None
    assert resource_no_id_no_type.get_type() is None

    



#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

# Placeholder for the generic parameterizable test suite for resources.
from typing import List

from kgforge.core import KnowledgeGraphForge, Resource
from kgforge.specializations.resources import Dataset


def test_from_resource(config, person, organization, store_metadata_value):
    forge = KnowledgeGraphForge(config)
    data = {
        'id': 'c51f4e4e-2b30-41f4-8f7c-aced85632b03',
        'type': ['Person', 'Agent'],
        'name': 'Jami Booth'
    }
    assert isinstance(person, Resource)
    dataset = Dataset.from_resource(forge, person)
    assert isinstance(dataset, Dataset)
    assert forge.as_json(dataset) == forge.as_json(person)
    assert forge.as_json(dataset) == data
    assert dataset._store_metadata is None

    person_with_store_metadata = Resource(**forge.as_json(person))
    person_with_store_metadata._store_metadata = store_metadata_value
    dataset = Dataset.from_resource(forge, person_with_store_metadata, store_metadata=True)
    assert isinstance(dataset, Dataset)
    person_with_store_metadata_json = forge.as_json(person_with_store_metadata)
    assert forge.as_json(dataset) == person_with_store_metadata_json
    assert forge.as_json(dataset) == data
    assert dataset._store_metadata == person_with_store_metadata._store_metadata
    assert forge.as_json(dataset, store_metadata=False) != forge.as_json(person_with_store_metadata, store_metadata=True)
    assert forge.as_json(dataset, store_metadata=True) == forge.as_json(person_with_store_metadata, store_metadata=True)

    assert isinstance(organization, Resource)
    dataset = Dataset.from_resource(forge, [person, organization])
    assert isinstance(dataset, List)
    assert len(dataset) == 2


#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

# Placeholder for the test suite for wrapping of property paths.

import pytest

from kgforge.core.wrappings.paths import Filter, create_filters_from_dict


@pytest.mark.parametrize(
    "dict_filters, expected",
    [
        pytest.param(
            ({"agent": {"name": "EPFL"}}),
            ([Filter(["agent", "name"], "__eq__", "EPFL")]),
            id="literal",
        ),
        pytest.param(
            ({"agent": {"name": "EPFL", "address": {"postalCode": "1015"}}}),
            (
                [
                    Filter(["agent", "name"], "__eq__", "EPFL"),
                    Filter(["agent", "address", "postalCode"], "__eq__", "1015"),
                ]
            ),
            id="one_nested",
        ),
        pytest.param(
            (
                {
                    "type": "Contribution",
                    "agent": {
                        "type": "Agent",
                        "name": "EPFL",
                        "address": {
                            "type": "PostalAddress",
                            "postalCode": "1015",
                            "addressCountry": {
                                "type": "Country",
                                "name": "Switzerland",
                            },
                        },
                    },
                }
            ),
            (
                [
                    Filter(["type"], "__eq__", "Contribution"),
                    Filter(["agent", "type"], "__eq__", "Agent"),
                    Filter(["agent", "name"], "__eq__", "EPFL"),
                    Filter(["agent", "address", "type"], "__eq__", "PostalAddress"),
                    Filter(["agent", "address", "postalCode"], "__eq__", "1015"),
                    Filter(
                        ["agent", "address", "addressCountry", "type"],
                        "__eq__",
                        "Country",
                    ),
                    Filter(
                        ["agent", "address", "addressCountry", "name"],
                        "__eq__",
                        "Switzerland",
                    ),
                ]
            ),
            id="many_nested",
        ),
    ],
)
def test_dict_to_filter(dict_filters, expected):
    filters = create_filters_from_dict(dict_filters)
    assert filters == expected


def test_bad_filter_operator():
    with pytest.raises(ValueError):
        Filter(["agent", "type"], "__equal__", "Agent")



#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

# Placeholder for the test suite for conversion of a resource to / from triples.
from copy import deepcopy

import os
from urllib.parse import urljoin
from urllib.request import pathname2url

import json
import pytest
from rdflib import Graph, BNode, term
from rdflib.namespace import RDF, Namespace

from kgforge.core import Resource
from kgforge.core.commons.exceptions import NotSupportedError
from kgforge.core.conversions.rdf import _merge_jsonld, from_jsonld, as_jsonld, Form, as_graph, from_graph, LD_KEYS

form_store_metadata_combinations = [
    pytest.param(Form.COMPACTED.value, True, id="compacted-with-metadata"),
    pytest.param(Form.COMPACTED.value, False, id="compacted-without-metadata"),
    pytest.param(Form.EXPANDED.value, True, id="expanded-with-metadata"),
    pytest.param(Form.EXPANDED.value, False, id="expanded-without-metadata")
]

store_metadata_params = [
    pytest.param(True, id="with-metadata"),
    pytest.param(False, id="without-metadata")
]


@pytest.fixture
def building_with_context(building, model_context):
    building.context = model_context.document["@context"]
    return building


class TestJsonLd:

    @pytest.mark.parametrize("store_metadata", store_metadata_params)
    def test_nested_resources(self, organization, organization_jsonld_compacted,
                              store_metadata, model_context, metadata_context):
        compacted = organization_jsonld_compacted(organization, store_metadata=store_metadata)
        result = as_jsonld(organization, form=Form.COMPACTED.value, store_metadata=store_metadata,
                           model_context=model_context, metadata_context=metadata_context, context_resolver=None)
        compacted["founder"]["@type"] = sorted(compacted["founder"]["@type"])
        result["founder"]["@type"] = sorted(result["founder"]["@type"])
        assert compacted == result

    @pytest.mark.parametrize("form, store_metadata", form_store_metadata_combinations)
    def test_no_context(self, form, store_metadata, valid_resource):
        with pytest.raises(NotSupportedError):
            as_jsonld(valid_resource, form=form, store_metadata=store_metadata, model_context=None,
                      metadata_context=None, context_resolver=None)

    @pytest.mark.parametrize("form, store_metadata", form_store_metadata_combinations)
    def test_unregistered_resource_with_context(self, building_with_context, building_jsonld,
                                                form, store_metadata, metadata_context):
        expected = building_jsonld(building_with_context, form, store_metadata, None)

        result = as_jsonld(building_with_context, form=form, store_metadata=store_metadata,
                           model_context=None, metadata_context=metadata_context,
                           context_resolver=None)
        assert expected == result

    @pytest.mark.parametrize("form, store_metadata", form_store_metadata_combinations)
    def test_registered_resource_with_context(self, building_with_context, make_registered,
                                              building_jsonld, form, store_metadata,
                                              store_metadata_value, metadata_context):
        base = urljoin('file:', pathname2url(os.getcwd()))
        registered_resource = make_registered(building_with_context, store_metadata_value, base)
        expected = building_jsonld(registered_resource, form, store_metadata, None)
        result = as_jsonld(registered_resource, form=form, store_metadata=store_metadata,
                           model_context=None, metadata_context=metadata_context,
                           context_resolver=None)
        assert expected == result

    @pytest.mark.parametrize("form, store_metadata", form_store_metadata_combinations)
    def test_unregistered_resource_model_context(self, building, model_context, building_jsonld,
                                                 form, store_metadata, metadata_context):
        expected = building_jsonld(building, form, store_metadata, model_context.document["@context"])
        result = as_jsonld(building, form=form, store_metadata=store_metadata,
                           model_context=model_context, metadata_context=metadata_context,
                           context_resolver=None)

        assert expected == result

    @pytest.mark.parametrize("form, store_metadata", form_store_metadata_combinations)
    def test_registered_resource_model_context(self, building_with_context, model_context,
                                               make_registered, building_jsonld, form,
                                               store_metadata, store_metadata_value,
                                               metadata_context):
        base = urljoin('file:', pathname2url(os.getcwd()))
        registered_resource = make_registered(building_with_context, store_metadata_value, base)
        expected = building_jsonld(registered_resource, form, store_metadata, None)
        result = as_jsonld(registered_resource, form=form, store_metadata=store_metadata,
                           model_context=model_context, metadata_context=metadata_context,
                           context_resolver=None)
        assert expected == result

    def test_from_jsonld(self, building, model_context, building_jsonld):
        building.context = model_context.document["@context"]
        payload = building_jsonld(building, "compacted", False, None)
        resource = from_jsonld(payload)
        assert resource == building

        payload_with_atvalue = deepcopy(payload)
        payload_with_atvalue["status"] = {
            "@type": "xsd:string",
            "@value": "opened"
        }
        resource_with_atvalue = from_jsonld(payload_with_atvalue)
        assert "value" not in LD_KEYS.keys()
        assert hasattr(resource_with_atvalue, "status")
        assert resource_with_atvalue.status == Resource.from_json({"type":"xsd:string", "@value":"opened"})

    def test_as_jsonld(self, building, model_context, building_jsonld, forge):
        building.context = model_context.document["@context"]
        building.context["embedding"] = {
            "@id": "http://example.org/embedding",
            "@container": "@list"
        }
        building.embedding = [0, 1, 2, 0]
        result = forge.as_jsonld(building, form="expanded")
        payload = building_jsonld(building, "expanded", False, None)
        payload["http://example.org/embedding"] = [{'@list': [
                                                             {"@value": 0},
                                                             {"@value": 1},
                                                             {"@value": 2},
                                                             {"@value": 0}
                                                             ]
                                                   }]

        assert result == payload

        building.embedding = [1, 2, 0, 0]
        building.a_list = [1, 9, 2, 0]
        building.a = []
        building.b = [None]
        building.c = [None, ""]
        building.d = ""
        building.e = None
        result = forge.as_jsonld(building, form="expanded")
        payload = building_jsonld(building, "expanded", False, None)
        payload["http://example.org/embedding"] = [{'@list': [
            {"@value": 1},
            {"@value": 2},
            {"@value": 0},
            {"@value": 0}
        ]
        }]
        payload["http://example.org/vocab/a"] = []
        payload["http://example.org/vocab/b"] = []
        payload["http://example.org/vocab/c"] = [{"@value": ""}]
        payload["http://example.org/vocab/d"] = [{"@value": ""}]
        payload["http://example.org/vocab/a_list"] = [{'@value': 1}, {'@value': 9}, {'@value': 2}, {'@value': 0}]

        assert result == payload

        result = forge.as_jsonld(building, form="compacted")
        payload = building_jsonld(building, "compacted", False, None)
        payload["embedding"] = [1,2,0,0]
        payload["a_list"] = [1, 9, 2, 0]
        payload["a"] = []
        payload["b"] = [None]
        payload["c"] = [None, ""]
        payload["d"] = ""
        payload["e"] = None

        assert result == payload


    def test_unresolvable_context(self, building, building_jsonld):
        building.context = "http://unresolvable.context.example.org/"
        payload = building_jsonld(building, "compacted", False, None)
        with pytest.raises(ValueError):
            from_jsonld(payload)


class TestGraph:

    @pytest.mark.parametrize("store_metadata", store_metadata_params)
    def test_as_graph(self, building, organization, building_jsonld, model_context,store_metadata, metadata_context):
        store_metadata = False
        building.id = "http://test/1234"
        building.context = model_context.document["@context"]
        data = building_jsonld(building, "compacted", store_metadata, None)
        expected = Graph().parse(data=json.dumps(data), format="json-ld")
        result = as_graph(building, store_metadata, model_context, None, None)
        _assert_same_graph(result, expected)

        organization_jsonld = as_jsonld(organization, form="compacted", store_metadata=store_metadata,
                  model_context=model_context, metadata_context=metadata_context,
                  context_resolver=None)
        expected = Graph()
        expected.parse(data=json.dumps(data), format="json-ld")
        expected.parse(data=json.dumps(organization_jsonld), format="json-ld")
        result = as_graph([building, organization], store_metadata, model_context, None, None)
        _assert_same_graph(result, expected)

    @pytest.mark.parametrize("store_metadata", store_metadata_params)
    def test_from_graph(self, building, organization, building_jsonld, model_context, store_metadata, metadata_context):
        store_metadata = False
        id = "http://test/1234"
        id_uri = term.URIRef(id)
        graph = Graph()
        graph.bind("schemaorg", Namespace("http://schema.org/"))
        graph.add((id_uri, RDF.type, term.URIRef("http://schema.org/Building")))
        graph.add((id_uri,term.URIRef("http://schema.org/name"),term.Literal("The Empire State Building")))
        graph.add((id_uri,term.URIRef("http://schema.org/description"),term.Literal("The Empire State Building is a 102-story landmark in New York City.")))
        graph.add((id_uri,term.URIRef("http://schema.org/image"),term.URIRef("http://www.civil.usherbrooke.ca/cours/gci215a/empire-state-building.jpg")))
        bNode = term.BNode()
        graph.add((id_uri,term.URIRef("http://schema.org/geo"),bNode))
        graph.add((bNode,term.URIRef("http://schema.org/latitude"), term.Literal("40.75")))
        results = from_graph(graph)

        assert isinstance(results, Resource)
        results.context.update({"schemaorg:latitude": {
                                  "@id": "http://schema.org/latitude",
                                  "@type": "xsd:float"
                                }})
        building.id = id
        building.context = model_context.document["@context"]
        expected = building_jsonld(building, "expanded", store_metadata, None)
        result_jsonld = as_jsonld(results, form="expanded", store_metadata=store_metadata,
                  model_context=model_context, metadata_context=metadata_context,
                  context_resolver=None)
        assert result_jsonld == expected

        graph.remove((id_uri, RDF.type, term.URIRef("http://schema.org/Building")))
        results = from_graph(graph)
        assert len(results) == 3

        graph.add((id_uri, RDF.type, term.URIRef("http://schema.org/Building")))
        graph.add((term.URIRef("http://www.civil.usherbrooke.ca/cours/gci215a/empire-state-building.jpg"), RDF.type, term.URIRef("http://schema.org/Image")))
        results = from_graph(graph, type=["http://schema.org/Building","http://schema.org/Image"])
        assert len(results) == 2
        assert results[0].type is not None
        assert results[1].type is not None

        result_0 = as_jsonld(results[0], form="expanded", store_metadata=store_metadata,
                  model_context=model_context, metadata_context=metadata_context,
                  context_resolver=None)

        result_1 = as_jsonld(results[1], form="expanded", store_metadata=store_metadata,
                             model_context=model_context, metadata_context=metadata_context,
                             context_resolver=None)
        results = [result_0, result_1]
        assert set(["http://schema.org/Building","http://schema.org/Image"]) == {result["@type"][0] for result in results}

        frame = {
            "@type": ['http://schema.org/Image'],
            "@embed": True
        }
        results = from_graph(graph, frame=frame)
        assert isinstance(results, Resource)
        expected = {'@type': ['http://schema.org/Image'], '@id': 'http://www.civil.usherbrooke.ca/cours/gci215a/empire-state-building.jpg'}
        result_jsonld = as_jsonld(results, form="expanded", store_metadata=store_metadata,
                         model_context=model_context, metadata_context=metadata_context,
                         context_resolver=None)
        assert result_jsonld == expected


def _assert_same_graph(result,expected):
    for s, p, o in expected:
        if isinstance(o, BNode):
            assert (s, p, None) in result
        elif isinstance(s, BNode):
            assert (None, p, o) in result
        else:
            assert (s, p, o) in result


class TestUtils:

    def test_merge_jsonld(self):
        str_ctx = "string_context"
        str_ctx2 = "string_context2"
        dict_ctx = {"a": 1, "b": 2}
        dict_ctx2 = {"c": 3, "d": 4}
        list_ctx = [str_ctx, dict_ctx]
        list_ctx2 = [str_ctx2, dict_ctx2]
        dict_ctx3 = {"a": 1, "b": 2, "c": 3, "d": 4}
        assert _merge_jsonld(str_ctx, str_ctx2) == [str_ctx, str_ctx2]
        assert _merge_jsonld(str_ctx, dict_ctx) == [str_ctx, dict_ctx]
        assert _merge_jsonld(str_ctx2, list_ctx) == [str_ctx2, str_ctx, dict_ctx]
        assert _merge_jsonld(list_ctx, str_ctx2) == [str_ctx, dict_ctx, str_ctx2]
        assert _merge_jsonld(list_ctx, dict_ctx2) == [str_ctx, dict_ctx3]
        assert _merge_jsonld(list_ctx, list_ctx2) == [str_ctx, dict_ctx3, str_ctx2]
        assert _merge_jsonld(dict_ctx, str_ctx) == [dict_ctx, str_ctx]
        assert _merge_jsonld(dict_ctx, list_ctx2) == [dict_ctx3, str_ctx2]
        assert _merge_jsonld(dict_ctx, dict_ctx2) == dict_ctx3
        assert _merge_jsonld(str_ctx, str_ctx) == str_ctx
        assert _merge_jsonld(dict_ctx, dict_ctx) == dict_ctx
        assert _merge_jsonld(list_ctx, list_ctx) == list_ctx
        assert _merge_jsonld(list_ctx, None) == list_ctx



#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.
from typing import Any, Optional, List

from abc import abstractmethod, ABC

from kgforge.core.commons.attributes import repr_class
from kgforge.core.commons.context import Context


class QueryBuilder(ABC):
    def __repr__(self) -> str:
        return repr_class(self)

    @staticmethod
    @abstractmethod
    def build(
        schema: Any,
        resolvers: Optional[List["Resolver"]],
        context: Context,
        *filters,
        **params
    ) -> Any:
        pass



# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.
from typing import List

import elasticsearch_dsl
import pytest
from elasticsearch_dsl import Text, Keyword, Nested, Object, Integer, Q


from kgforge.core.commons.es_query_builder import (
    _recursive_resolve_nested,
    _build_keyword_path,
    _detect_mapping_type,
    ESQueryBuilder,
)
from kgforge.core.wrappings import Filter


class TestESQueryBuilder:
    @pytest.mark.parametrize(
        "property_path, mapping_type, dynamic_mapping_type, default_str_keyword_field, keyword_path",
        [
            pytest.param(
                ("annotation.hasBody.label"),
                (Text(fields={"keyword": Keyword()})),
                (None),
                (None),
                ("annotation.hasBody.label.keyword"),
                id="text_keyword_field",
            ),
            pytest.param(
                ("@type"),
                (Keyword()),
                (None),
                (None),
                ("@type"),
                id="keyword_field_path",
            ),
            pytest.param(
                ("objectOfStudy.label"),
                (Text()),
                (None),
                (None),
                (None),
                id="non_keyword_field",
            ),
            pytest.param(
                ("brainLocation.brainRegion"),
                (Object()),
                (None),
                (None),
                (None),
                id="object_field_path",
            ),
            pytest.param(
                ("contribution.agent.type"),
                (None),
                (Text()),
                ("keyword"),
                ("contribution.agent.type.keyword"),
                id="dynamic_mapping_type",
            ),
            pytest.param(
                ("contribution.agent.type"),
                (None),
                (None),
                ("keyword"),
                (None),
                id="noat_mapping_type",
            ),
        ],
    )
    def test__build_keyword_path(
        self,
        property_path,
        mapping_type,
        dynamic_mapping_type,
        default_str_keyword_field,
        keyword_path,
    ):
        k_p = _build_keyword_path(
            mapping_type, property_path, dynamic_mapping_type, default_str_keyword_field
        )
        assert k_p == keyword_path

    @pytest.mark.parametrize(
        "field_path, nested_path, found_path, mapping_type",
        [
            pytest.param(
                (["annotation", "hasBody", "label"]),
                (["annotation.hasBody"]),
                (["annotation","hasBody", "label"]),
                (Text(fields={"keyword": Keyword()})),
                id="text_keyword_field_in_nested",
            ),
            pytest.param(
                (["annotation", "hasBody"]),
                (["annotation.hasBody"]),
                (["annotation","hasBody"]),
                (Nested()),
                id="nested_field",
            ),
            pytest.param(
                (["annotation", "hasBody", "isMeasurementOf", "label"]),
                (["annotation.hasBody"]),
                (["annotation","hasBody"]),
                (Nested()),
                id="non_existing_field_in_nested",
            ),
            pytest.param(
                (["brainLocation", "brainRegion"]),
                ([]),
                (["brainLocation","brainRegion"]),
                (Object()),
                id="object_field"
            ),
            pytest.param(
                (["brainLocation", "brainRegion", "label"]),
                ([]),
                (["brainLocation", "brainRegion", "label"]),
                (Text(fields={"keyword": Keyword()})),
                id="text_keyword_field_in_object",
            ),
            pytest.param(
                (["an_integer"]),
                ([]),
                (["an_integer"]),
                (Integer(fields={"keyword": Keyword()})),
                id="integer_field",
            ),
            pytest.param((["@type"]), ([]), (["@type"]), (Keyword()), id="keyword_field"),
        ],
    )
    def test__recursive_resolve_nested(
        self, field_path, nested_path, found_path, mapping_type, es_mapping_dict
    ):
        es_mapping = elasticsearch_dsl.Mapping()
        es_mapping._update_from_dict(es_mapping_dict)
        n_p, f_p, m_t = _recursive_resolve_nested(es_mapping, field_path)
        assert n_p == nested_path
        assert f_p == found_path
        assert str(m_t) == str(mapping_type)

    @pytest.mark.parametrize(
        "value, mapping_type",
        [
            pytest.param(
                ("A str value"), (elasticsearch_dsl.Text()), id="text_mapping_type"
            ),
            pytest.param(
                ("2020-10-20T13:53:22.880Z"),
                (elasticsearch_dsl.Date()),
                id="date_mapping_type",
            ),
            pytest.param((3), (elasticsearch_dsl.Long()), id="integer_mapping_type"),
            pytest.param(
                ("8.2"), (elasticsearch_dsl.Float()), id="str_double_mapping_type"
            ),
            pytest.param(
                (True), (elasticsearch_dsl.Boolean()), id="true_bool_mapping_type"
            ),
            pytest.param(
                (False), (elasticsearch_dsl.Boolean()), id="false_bool_mapping_type"
            ),
        ],
    )
    def test__detect_mapping_type(self, value, mapping_type):
        m_t = _detect_mapping_type(value)
        assert str(m_t) == str(mapping_type)

    @pytest.mark.parametrize(
        "filters, params, es_query",
        [
            pytest.param(
                (
                    [
                        Filter(
                            operator="__eq__",
                            path=["annotation", "hasBody", "label"],
                            value="An annotation",
                        )
                    ]
                ),
                ({}),
                (
                    elasticsearch_dsl.Search().filter(
                        "nested",
                        path="annotation.hasBody",
                        query=elasticsearch_dsl.query.Bool(
                            filter=[
                                elasticsearch_dsl.query.Q(
                                    "term",
                                    **{
                                        "annotation.hasBody.label.keyword": "An annotation"
                                    },
                                )
                            ]
                        ),
                    )
                ),
                id="build_nested_filter_query",
            ),
            pytest.param(
                (
                    [
                        Filter(
                            operator="__eq__",
                            path=["annotation", "hasBody", "isMeasurementOf"],
                            value="An annotation",
                        )
                    ]
                ),
                ({}),
                (
                    elasticsearch_dsl.Search().query(
                        "bool",
                        must=[
                            elasticsearch_dsl.query.Nested(
                                path="annotation.hasBody",
                                query=elasticsearch_dsl.query.Bool(
                                    must=[
                                        elasticsearch_dsl.query.Q(
                                            "match",
                                            **{
                                                "annotation.hasBody.isMeasurementOf": "An annotation"
                                            },
                                        )
                                    ]
                                ),
                            ).to_dict()
                        ],
                    )
                    # )
                ),
                id="build_unknown_nested_must_query",
            ),
            pytest.param(
                (
                    [
                        Filter(
                            operator="__eq__",
                            path=["brainLocation", "brainRegion", "label"],
                            value="A label",
                        )
                    ]
                ),
                ({}),
                (
                    elasticsearch_dsl.Search().query(
                        elasticsearch_dsl.query.Bool(
                            filter=[
                                elasticsearch_dsl.query.Term(
                                    **{
                                        "brainLocation.brainRegion.label.keyword": "A label"
                                    }
                                )
                            ]
                        )
                    )
                ),
                id="build_filter_query",
            ),
            pytest.param(
                (
                    [
                        Filter(
                            operator="__eq__",
                            path=["brainLocation", "layer", "label"],
                            value="A label",
                        )
                    ]
                ),
                ({}),
                (
                    elasticsearch_dsl.Search().query(
                        elasticsearch_dsl.query.Bool(
                            must=[
                                elasticsearch_dsl.query.Match(
                                    **{"brainLocation.layer.label": "A label"}
                                )
                            ]
                        )
                    )
                ),
                id="build_match_query",
            ),
            pytest.param(
                (
                    [
                        Filter(
                            operator="__eq__",
                            path=["brainLocation", "brainRegion", "isDefinedBy"],
                            value="A definition",
                        )
                    ]
                ),
                ({}),
                (
                    elasticsearch_dsl.Search().query(
                        elasticsearch_dsl.query.Bool(
                            filter=[
                                elasticsearch_dsl.query.Term(
                                    **{
                                        "brainLocation.brainRegion.isDefinedBy.keyword": "A definition"
                                    }
                                )
                            ]
                        )
                    )
                ),
                id="build_unknown_filter_query",
            ),
            pytest.param(
                (
                    [
                        Filter(
                            operator="__ne__",
                            path=["brainLocation", "brainRegion", "label"],
                            value="A label",
                        )
                    ]
                ),
                ({}),
                (
                    elasticsearch_dsl.Search().query(
                        elasticsearch_dsl.query.Bool(
                            must_not=[
                                elasticsearch_dsl.query.Term(
                                    **{
                                        "brainLocation.brainRegion.label.keyword": "A label"
                                    }
                                )
                            ]
                        )
                    )
                ),
                id="build_must_not_filter_query",
            ),
            pytest.param(
                (
                    [
                        Filter(
                            operator="__ne__",
                            path=["brainLocation", "layer", "label"],
                            value="A label",
                        )
                    ]
                ),
                ({}),
                (
                    elasticsearch_dsl.Search().query(
                        elasticsearch_dsl.query.Bool(
                            must_not=[
                                elasticsearch_dsl.query.Match(
                                    **{"brainLocation.layer.label": "A label"}
                                )
                            ]
                        )
                    )
                ),
                id="build_must_not_match_query",
            ),
            pytest.param(
                ([Filter(operator="__gt__", path=["an_integer"], value="3")]),
                ({}),
                (
                    elasticsearch_dsl.Search().query(
                        elasticsearch_dsl.query.Bool(
                            filter=[
                                elasticsearch_dsl.query.Range(
                                    **{"an_integer": {"gt": "3"}}
                                )
                            ]
                        )
                    )
                ),
                id="build_range_gt_query",
            ),
            pytest.param(
                ([Filter(operator="__le__", path=["series", "a_float"], value="2.0")]),
                ({}),
                (
                    elasticsearch_dsl.Search().filter(
                        "nested",
                        path="series",
                        query=elasticsearch_dsl.query.Bool(
                            filter=[
                                elasticsearch_dsl.query.Range(
                                    **{"series.a_float": {"lte": "2.0"}}
                                )
                            ]
                        ),
                    )
                ),
                id="build_range_le_nested_query",
            ),
            pytest.param(
                (
                    [
                        Filter(
                            operator="__eq__",
                            path=["unknown", "path"],
                            value="Unknown value",
                        )
                    ]
                ),
                ({"default_str_keyword_field": "keyword"}),
                (
                    elasticsearch_dsl.Search().query(
                        elasticsearch_dsl.query.Bool(
                            filter=[
                                elasticsearch_dsl.query.Term(
                                    **{"unknown.path.keyword": "Unknown value"}
                                )
                            ]
                        )
                    )
                ),
                id="build_complete_unknown_path_filter_text",
            ),
            pytest.param(
                (
                    [
                        Filter(
                            operator="__eq__",
                            path=["unknown", "path"],
                            value="Unknown value",
                        )
                    ]
                ),
                ({}),
                (
                    elasticsearch_dsl.Search().query(
                        elasticsearch_dsl.query.Bool(
                            filter=[
                                elasticsearch_dsl.query.Term(
                                    **{"unknown.path.keyword": "Unknown value"}
                                )
                            ]
                        )
                    )
                ),
                id="build_complete_unknown_path_filter_text_no_default_str_keyword_field",
            ),
            pytest.param(
                ([Filter(operator="__eq__", path=["unknown", "path"], value="3.0")]),
                ({}),
                (
                    elasticsearch_dsl.Search().query(
                        elasticsearch_dsl.query.Bool(
                            must=[
                                elasticsearch_dsl.query.Match(**{"unknown.path": "3.0"})
                            ]
                        )
                    )
                ),
                id="build_complete_unknown_path_must_match_float",
            ),
            pytest.param(
                (
                    [
                        Filter(
                            operator="__eq__",
                            path=["a_dense_vector"],
                            value=[0.1, 0.3, 0.8],
                        )
                    ]
                ),
                ({}),
                (
                    elasticsearch_dsl.Search().query(
                        elasticsearch_dsl.query.ScriptScore(
                            query=elasticsearch_dsl.query.Bool(
                                must=[
                                    elasticsearch_dsl.query.Exists(
                                        field="a_dense_vector"
                                    )
                                ]
                            ),
                            script={
                                "source": "doc['a_dense_vector'].size() == 0 ? 0 : (cosineSimilarity(params.queryVector, doc['a_dense_vector'])+1.0) / 2",
                                "params": {"queryVector": [0.1, 0.3, 0.8]},
                            },
                        )
                    )
                ),
                id="build_dense_vector_query",
            ),
            pytest.param(
                (
                    [
                        Filter(
                            operator="__eq__",
                            path=["derivation", "a_dense_vector"],
                            value=[0.1, -0.3, 0.8],
                        )
                    ]
                ),
                ({}),
                (
                    elasticsearch_dsl.Search().query(
                        elasticsearch_dsl.query.Nested(
                            path="derivation",
                            query=elasticsearch_dsl.query.ScriptScore(
                                query=elasticsearch_dsl.query.Bool(
                                    must=[
                                        elasticsearch_dsl.query.Exists(
                                            field="derivation.a_dense_vector"
                                        )
                                    ]
                                ),
                                script={
                                    "source": "doc['derivation.a_dense_vector'].size() == 0 ? 0 : (cosineSimilarity(params.queryVector, doc['derivation.a_dense_vector'])+1.0) / 2",
                                    "params": {"queryVector": [0.1, -0.3, 0.8]},
                                },
                            ),
                        )
                    )
                ),
                id="build_nested_dense_vector_query",
            ),
            pytest.param(
                (
                    [
                        Filter(
                            operator="__eq__",
                            path=["brainLocation", "brainRegion", "label"],
                            value="A label",
                        )
                    ]
                ),
                ({"includes": ["to_include"], "excludes": ["to_exclude"]}),
                (
                    elasticsearch_dsl.Search()
                    .query(
                        elasticsearch_dsl.query.Bool(
                            filter=[
                                elasticsearch_dsl.query.Term(
                                    **{
                                        "brainLocation.brainRegion.label.keyword": "A label"
                                    }
                                )
                            ]
                        )
                    )
                    .source(includes=["to_include"], excludes=["to_exclude"])
                ),
                id="build_query_with_source",
            ),
            pytest.param(
                (
                    [
                        Filter(
                            operator="__eq__",
                            path=["type", "id"],
                            value="Dataset",
                        ),
                        Filter(
                            operator="__eq__",
                            path=["contribution", "agent", "type"],
                            value="Person",
                        ),
                        Filter(
                            operator="__eq__",
                            path=["contribution", "agent", "id"],
                            value="https://orcid",
                        ),
                        Filter(
                            operator="__ne__",
                            path=["derivation", "entity", "@type"],
                            value="Entity",
                        ),
                    ]
                ),
                ({}),
                (
                    elasticsearch_dsl.Search().query(
                        "bool",
                        filter=[
                            elasticsearch_dsl.query.Term(
                                **{
                                    "@type": "Dataset"
                                }
                            ).to_dict(),
                            elasticsearch_dsl.query.Nested(
                                path="contribution.agent",
                                query=elasticsearch_dsl.query.Bool(
                                    filter=[
                                        elasticsearch_dsl.query.Term(
                                            **{
                                                "contribution.agent.@type.keyword": "Person"
                                            }
                                        )
                                    ]
                                ),
                            ).to_dict(),
                            elasticsearch_dsl.query.Nested(
                                path="contribution.agent",
                                query=elasticsearch_dsl.query.Bool(
                                    filter=[
                                        elasticsearch_dsl.query.Term(
                                            **{
                                                "contribution.agent.@id": "https://orcid"
                                            }
                                        )
                                    ]
                                ),
                            ).to_dict(),
                        ],
                        must_not=[
                            elasticsearch_dsl.query.Nested(
                                path="derivation.entity",
                                query=elasticsearch_dsl.query.Bool(
                                    must_not=[
                                        elasticsearch_dsl.query.Term(
                                            **{
                                                "derivation.entity.@type.keyword": "Entity"
                                            }
                                        )
                                    ]
                                ),
                            ).to_dict()
                        ],
                    )
                ),
                id="rewrite_type_id",
            ),
            pytest.param(
                (
                    [
                        Filter(
                            operator="__eq__",
                            path=["brainLocation", "brainRegion", "label"],
                            value=["A label", "Another label"],
                        )
                    ]
                ),
                ({}),
                (
                    elasticsearch_dsl.Search()
                    .query(
                        elasticsearch_dsl.query.Bool(
                            filter=[
                                elasticsearch_dsl.query.Terms(
                                    **{
                                        "brainLocation.brainRegion.label.keyword": ["A label", "Another label"]
                                    }
                                )
                            ]
                        )
                    )
                ),
                id="build_query_with_multiple_values",
            )
        ],
    )
    def test_build(self, filters, params, es_query, es_mapping_dict):
        es_mapping = elasticsearch_dsl.Mapping()
        es_mapping._update_from_dict(es_mapping_dict)
        query = ESQueryBuilder.build(es_mapping_dict, None, None, filters, **params)
        assert isinstance(query, dict)
        assert elasticsearch_dsl.Search.from_dict(query) == es_query

    @pytest.mark.parametrize(
        "filters, default_str_keyword_field",
        [
            pytest.param(
                (
                    [
                        Filter(
                            operator="__eq__",
                            path=["brainLocation", "brainRegion"],
                            value="A brainRegion",
                        )
                    ]
                ),
                (None),
                id="build_failing_object_query",
            ),
            pytest.param(
                (
                    [
                        Filter(
                            operator="__gt__",
                            path=["brainLocation", "brainRegion", "label"],
                            value="A label",
                        )
                    ]
                ),
                (None),
                id="build_failing_gt_Text_query",
            ),
            pytest.param(
                (
                    [
                        Filter(
                            operator="__ge__",
                            path=["annotation", "hasBody", "label"],
                            value="A label",
                        )
                    ]
                ),
                (None),
                id="build_failing_ge_nested_Text_query",
            ),
            pytest.param(
                ([Filter(operator="__lt__", path=["a_boolean"], value="A boolean")]),
                (None),
                id="build_failing_lt_Boolean_query",
            ),
            pytest.param(
                (
                    [
                        Filter(
                            operator="__lt__",
                            path=["a_dense_vector"],
                            value=[0.1, 0.3, 0.8],
                        )
                    ]
                ),
                (None),
                id="build_failing_dense_vector_query",
            ),
            pytest.param(
                (
                    [
                        Filter(
                            operator="__eq__",
                            path=["a_dense_vector"],
                            value=[0.1, 0.3, 0.8],
                        ),
                        Filter(
                            operator="__eq__",
                            path=["a_dense_vector"],
                            value=[0.5, 0.5, 0.9],
                        ),
                    ]
                ),
                (None),
                id="build_failing_multiple_dense_vector_query",
            ),
        ],
    )
    def test_build_exception(self, filters, default_str_keyword_field, es_mapping_dict):
        es_mapping = elasticsearch_dsl.Mapping()
        es_mapping._update_from_dict(es_mapping_dict)

        with pytest.raises(ValueError):
            ESQueryBuilder.build(
                es_mapping_dict,
                None,
                None,
                filters,
                default_str_keyword_field=default_str_keyword_field,
            )

    @pytest.mark.parametrize(
        "filters, default_str_keyword_field",
        [
            pytest.param(
                (
                    [
                        Filter(
                            operator="__eq__",
                            path=["annotation", "hasBody", "isMeasurementOf"],
                            value="An annotation",
                        )
                    ]
                ),
                (None),
                id="build_nested_filter_query",
            )
        ],
    )
    def test_build_dynamic_false(
        self, filters, default_str_keyword_field, es_mapping_dict
    ):
        es_mapping = elasticsearch_dsl.Mapping()
        es_mapping._update_from_dict(es_mapping_dict)

        with pytest.raises(ValueError):
            es_mapping_dict["dynamic"] = False
            ESQueryBuilder.build(
                es_mapping_dict,
                None,
                None,
                filters,
                default_str_keyword_field=default_str_keyword_field,
            )
